package org.artifactory.event;

import com.google.common.annotations.VisibleForTesting;
import org.apache.commons.lang3.concurrent.BasicThreadFactory;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.common.ConstantValues;
import org.artifactory.queue.execution.QueueTriggeringService;
import org.artifactory.spring.ContextCreationListener;
import org.artifactory.storage.db.security.service.BasicCacheModel;
import org.artifactory.storage.db.security.service.VersionCache;
import org.artifactory.storage.db.security.service.VersioningCacheImpl;
import org.artifactory.storage.event.EventsService;
import org.artifactory.storage.event.ShiftEventsService;
import org.jfrog.common.concurrency.ConflictGuard;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;

import javax.annotation.PreDestroy;
import java.util.Set;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * The thread will sleep 30 sec or until will signalled for shifts
 * If 30 sec elapsed or got shift triggering (version promoted), a shift will invoked.
 * Shift may be locked for max 120 sec trying to get distributed lock.
 * If distributed lock was not acquired, a new shift will be invoked right away.
 *
 * @author shaharz
 */

@Service
public class ShiftEventsServiceImpl implements Runnable, ShiftEventsService, ContextCreationListener {

    private static final Logger log = LoggerFactory.getLogger(ShiftEventsServiceImpl.class);

    public static final String SHIFT_EVENTS_CONFLICT_GUARD = "shift-events";
    private static final long TRY_LOCK_TIMEOUT_MS = ConstantValues.shiftEventsTryLockTimeoutMs.getInt();
    private static final long PERIODIC_SHIFT_MS = ConstantValues.shiftEventsPeriodicMs.getInt();
    private static final BasicCacheModel CACHE_MODEL = new BasicCacheModel() {
        @Override
        public long getVersion() {
            return 0;
        }

        @Override
        public void setVersion(long version) {
        }

        @Override
        public void destroy() {
        }
    };

    private final ConflictGuard conflictGuard;
    private volatile boolean run = true;

    @VisibleForTesting
    final ThreadPoolExecutor executor;
    @VisibleForTesting
    final EventsService eventsService;
    @VisibleForTesting
    final QueueTriggeringService queueTriggeringService;
    @VisibleForTesting
    final VersionCache<?> versionCache;
    @VisibleForTesting
    final Object lock = new Object();

    public ShiftEventsServiceImpl(AddonsManager addonsManager, EventsService eventsService,
            QueueTriggeringService queueTriggeringService) {
        this.conflictGuard = addonsManager.addonByType(HaCommonAddon.class)
                .getConflictGuard(SHIFT_EVENTS_CONFLICT_GUARD);

        this.eventsService = eventsService;
        this.queueTriggeringService = queueTriggeringService;
        this.executor = new ThreadPoolExecutor(1, 1, 0L, TimeUnit.MILLISECONDS,
                new LinkedBlockingQueue<>(), new BasicThreadFactory.Builder()
                .namingPattern(ShiftEventsServiceImpl.class.getSimpleName() + "-%1$s").build());
        this.versionCache = new VersioningCacheImpl(TRY_LOCK_TIMEOUT_MS, () -> {
            shiftEvents();
            return CACHE_MODEL;
        }, this.conflictGuard);
    }

    @Override
    public void onContextCreated() {
        start();
    }

    @Override
    public void run() {
        log.debug("Thread started...");
        while (run) {
            try {
                versionCache.get();
                if (run) {
                    synchronized (lock) {
                        // if there is no new version:
                        // it means that the shift was not executed due to try lock timeout
                        // in that case we don't want to sleep for 30 sec, we want to try shift again
                        if (!versionCache.hasNewVersion()) {
                            await();
                            // promote teh version after sleep, so it will try to shift
                            versionCache.promoteVersion();
                        }
                    }
                }
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            } catch (Exception e) {
                log.error("Unexpected error", e);
            }
        }
        log.debug("Thread ended.");
    }

    @Override
    public void triggerShift() {
        synchronized (lock) {
            int version = versionCache.promoteVersion();
            log.debug("Shift events triggered, version: {}", version);
            lock.notifyAll();
        }
    }

    @PreDestroy
    void stop() throws InterruptedException {
        synchronized (lock) {
            run = false;
            lock.notifyAll();
        }
        executor.shutdown();
        log.info("Shutdown terminated: {}", executor.awaitTermination(5, TimeUnit.SECONDS));
        log.debug("Stopped.");
    }

    @VisibleForTesting
    void shiftEvents() {
        log.debug("Shifting events...");
        // Copy the events from tm tmp to node_events and clear the tmp table in one TX.
        while (true) {
            Set<String> reposKey = eventsService.shiftEvents();
            if (!reposKey.isEmpty()) {
                // trigger the queues for the effected repos key
                queueTriggeringService.triggerQueues(reposKey);
                log.debug("Queues triggered: {}", reposKey);
            } else {
                break;
            }
        }
    }

    @VisibleForTesting
    void await() throws InterruptedException {
        synchronized (lock) {
            log.debug("About to wait {} ms.", PERIODIC_SHIFT_MS);
            lock.wait(PERIODIC_SHIFT_MS);
        }
    }

    private void start() {
        synchronized (lock) {
            if (executor.getActiveCount() == 0) {
                executor.submit(this);
                log.debug("Started.");
            }
        }
    }
}
