package org.artifactory.projects;

import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import org.apache.commons.lang.StringUtils;
import org.apache.http.HttpStatus;
import org.artifactory.addon.build.BuildAddon;
import org.artifactory.common.MutableStatusHolder;
import org.artifactory.util.AolUtils;
import org.artifactory.util.CollectionUtils;
import org.slf4j.Logger;

import javax.annotation.Nullable;
import javax.validation.constraints.NotNull;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**
 * @author BarakH
 */
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class ProjectsUtils {

    public static final String PROJ_LIMIT_MSG = "Project storage limit has been reached. Contact your project" +
            " administrator to change the storage quota limits";
    private static final Pattern PROJECT_REPO_KEY_PATTERN = Pattern.compile("^([a-zA-Z0-9]*)-(.*)");

    public static boolean isDefaultProject(@Nullable String project) {
        return StringUtils.isEmpty(project) || ProjectsService.DEFAULT_PROJECT.equalsIgnoreCase(project);
    }

    /**
     * The method checks if the provided projectKey representing 'Unassigned' context (i.e resources that are not
     * assigned to any project)
     */
    public static boolean isUnassignedContext(@Nullable String projectKey) {
        return StringUtils.isEmpty(projectKey) || ProjectsService.UNASSIGNED.equalsIgnoreCase(projectKey);
    }

    /**
     * @throws IllegalArgumentException if the given project does not contain all the virtual repositories
     */
    public static void assertVirtualRepositoriesAreInProject(@NonNull ProjectsService projectsService,
            @Nullable String projectKey, @Nullable List<String> virtualReposList) throws IllegalArgumentException {
        if (isDefaultProject(projectKey) || CollectionUtils.isNullOrEmpty(virtualReposList)) {
            return;
        }
        Set<String> projectRepositories = projectsService.getProjectRepositoriesKeys(projectKey, false);
        if (!projectRepositories.containsAll(virtualReposList)) {
            throw new IllegalArgumentException(
                    "Virtual repository must include only repositories from the same project it was created with");
        }
    }

    @Nullable
    public static String getProjectKeyFromRepoKey(@NonNull String repoKey) {
        if (Objects.equals(repoKey, BuildAddon.BUILD_INFO_REPO_NAME)) {
            return ProjectsService.DEFAULT_PROJECT;
        }

        Matcher matcher = PROJECT_REPO_KEY_PATTERN.matcher(repoKey);
        if (matcher.matches()) {
            return matcher.group(1);
        }
        return null;
    }

    @NotNull
    public static String getRepoKeyFromProjectKey(@NonNull String projectKey) {
        return Objects.equals(projectKey, ProjectsService.DEFAULT_PROJECT) ?
                BuildAddon.BUILD_INFO_REPO_NAME : projectKey + BuildAddon.BUILD_INFO_REPO_EXTENSION;
    }

    /**
     * Indicating whether to use project storage quota SOFT limit or not
     * always true if in AOL environment
     *
     * @param isSoftLimitConfiguredValue - the value taken from project entity
     */
    public static boolean isStorageQuotaSoftLimit(@NonNull boolean isSoftLimitConfiguredValue) {
        return isSoftLimitConfiguredValue || AolUtils.isAol();
    }

    public static boolean handleStatusUpdateByQuotaLimitType(MutableStatusHolder statusHolder, boolean isSoftLimit,
            Logger log) {
        if (ProjectsUtils.isStorageQuotaSoftLimit(isSoftLimit)) {
            statusHolder.warn(PROJ_LIMIT_MSG, HttpStatus.SC_REQUEST_TOO_LONG, log);
            return false;
        } else {
            statusHolder.error(PROJ_LIMIT_MSG, HttpStatus.SC_REQUEST_TOO_LONG, log);
            return true;
        }
    }
}
