package org.artifactory.build;

import org.artifactory.addon.build.BuildAddon;
import org.artifactory.api.rest.build.ContinueBuildFilter;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.storage.db.build.entity.BuildIdEntity;
import org.artifactory.storage.db.build.service.BuildIdImpl;
import org.glassfish.jersey.internal.guava.Predicates;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;

/**
 * @author Omri Ziv
 */

@Test
public class BuildServiceUtilsTest {

    private final long now = System.currentTimeMillis();

    @Mock
    ArtifactoryHome artifactoryHome;
    @Mock
    ArtifactorySystemProperties artifactorySystemProperties;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.initMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        when(artifactorySystemProperties.getLongProperty(any(ConstantValues.class)))
                .thenAnswer(invocationOnMock -> {
                    ConstantValues constantValue = invocationOnMock.getArgument(0);
                    return Long.parseLong(constantValue.getDefValue().trim());
                });
    }

    @AfterMethod
    public void cleanup() {
        ArtifactoryHome.unbind();
    }

    public void testIterateFetchFromDb() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        List<BuildIdEntity> list = BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, Predicates.alwaysTrue(),
                buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                        buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(), buildIdEntity.getBuildRepo()),
                continueBuildFilter);
        assertEquals(10, list.size());

    }

    public void testIterateFetchFromDb0Limit() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(0L);
        List<BuildIdEntity> list = BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, Predicates.alwaysTrue(),
                buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                        buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(), buildIdEntity.getBuildRepo()),
                continueBuildFilter);
        assertEquals(0, list.size());
    }

    public void testIterateFetchFromDbNoResults() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(0L);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(x -> Collections.emptyList(), Predicates.alwaysTrue(),
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(0, list.size());
    }


    public void testIterateFetchFromDbFilterOdds() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, build -> build.getBuildId() % 2 == 0,
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(10, list.size());
        assertEquals(2, list.get(0).getBuildId());
        assertEquals(20, list.get(9).getBuildId());

    }

    public void testIterateFetchFromDbFilterOddsStartFrom() {
        long now = System.currentTimeMillis();
        ContinueBuildFilter continueBuildFilter = new ContinueBuildFilter();
        continueBuildFilter.setLimit(10L);
        BuildId continueBuildId = new BuildIdImpl(16, "build16", "16", now, BuildAddon.BUILD_INFO_REPO_NAME);
        continueBuildFilter.setContinueBuildId(continueBuildId);
        List<BuildIdEntity> list =
                BuildServiceUtils.iterateFetchFromDb(this::fetchFunction1, build -> build.getBuildId() % 2 == 0,
                        buildIdEntity -> new BuildIdImpl(buildIdEntity.getBuildId(), buildIdEntity.getBuildName(),
                                buildIdEntity.getBuildNumber(), buildIdEntity.getBuildDate(),
                                buildIdEntity.getBuildRepo()),
                        continueBuildFilter);
        assertEquals(10, list.size());
        assertEquals(18, list.get(0).getBuildId());
        assertEquals(36, list.get(9).getBuildId());

    }

    public void test() {
        BuildCoordinate coordinate =
                BuildServiceUtils.extractCoordinatesFromPath("yevdo-build/2-1592480449869.json",
                        BuildAddon.BUILD_INFO_REPO_NAME);
        assertNotNull(coordinate);
        assertEquals("yevdo-build", coordinate.getBuildName());
        assertEquals("2", coordinate.getBuildNumber());
        assertEquals(1592480449869L, coordinate.getDate());

        coordinate = BuildServiceUtils
                .extractCoordinatesFromPath("artifactory-build-info:blabla.json", BuildAddon.BUILD_INFO_REPO_NAME);
        assertNull(coordinate);
    }

    private List<BuildIdEntity> fetchFunction1(ContinueBuildFilter continueBuildFilter) {
        return IntStream.rangeClosed(1, 100)
                .mapToObj(number -> new BuildIdEntity(number, "build-" + number, number + "", now,
                        BuildAddon.BUILD_INFO_REPO_NAME))
                .filter(buildIdEntity -> {
                    return continueBuildFilter.getContinueBuildId() == null ||
                            Integer.parseInt(buildIdEntity.getBuildNumber()) >
                                    Integer.parseInt(continueBuildFilter.getContinueBuildId().getNumber());
                })
                .limit(continueBuildFilter.getLimit())
                .collect(Collectors.toList());
    }

}
