package org.artifactory.event;

import ch.qos.logback.classic.Level;
import ch.qos.logback.classic.Logger;
import com.google.common.collect.Sets;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.ha.HaCommonAddon;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.common.property.ArtifactorySystemProperties;
import org.artifactory.queue.execution.QueueTriggeringService;
import org.artifactory.storage.event.EventsService;
import org.awaitility.Awaitility;
import org.jfrog.common.concurrency.ConflictGuard;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;
import org.mockito.stubbing.Answer;
import org.slf4j.LoggerFactory;
import org.testng.Assert;
import org.testng.annotations.AfterClass;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeClass;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.Test;

import java.time.Duration;
import java.time.Instant;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

@Test
public class ShiftEventsServiceImplTest {

    private ShiftEventsServiceImpl shiftEventsService;

    @Mock
    private ArtifactoryHome artifactoryHome;
    @Mock
    private AddonsManager addonsManager;
    @Mock
    private HaCommonAddon haCommonAddon;
    @Mock
    private ConflictGuard conflictGuard;
    @Mock
    private ArtifactorySystemProperties artifactorySystemProperties;
    @Mock
    private EventsService eventsService;
    @Mock
    private QueueTriggeringService queueTriggeringService;

    private AtomicInteger aWaits = null;

    @BeforeClass
    private void before() {
        ((Logger) LoggerFactory.getILoggerFactory().getLogger(ShiftEventsServiceImpl.class.getName()))
                .setLevel(Level.DEBUG);
    }

    @AfterClass
    private void after() {
        ((Logger) LoggerFactory.getILoggerFactory().getLogger(ShiftEventsServiceImpl.class.getName()))
                .setLevel(Level.INFO);
    }

    @BeforeMethod
    private void setup() throws Exception {
        MockitoAnnotations.openMocks(this);
        ArtifactoryHome.bind(artifactoryHome);
        Mockito.when(artifactoryHome.getArtifactoryProperties()).thenReturn(artifactorySystemProperties);
        Mockito.when(
                artifactorySystemProperties.getLongProperty(Mockito.eq(ConstantValues.shiftEventsTryLockTimeoutMs))).
                thenReturn(5000L);
        Mockito.when(artifactorySystemProperties.getLongProperty(Mockito.eq(ConstantValues.shiftEventsPeriodicMs))).
                thenReturn(30000L);
        Mockito.when(addonsManager.addonByType(Mockito.eq(HaCommonAddon.class))).thenReturn(haCommonAddon);
        Mockito.when(haCommonAddon.getConflictGuard(Mockito.eq(ShiftEventsServiceImpl.SHIFT_EVENTS_CONFLICT_GUARD)))
                .thenReturn(conflictGuard);

        Mockito.when(conflictGuard.tryToLock(Mockito.anyLong(), Mockito.any(TimeUnit.class))).thenReturn(true);

        aWaits = new AtomicInteger();
        shiftEventsService = new ShiftEventsServiceImpl(addonsManager, eventsService, queueTriggeringService) {
            @Override
            void await() throws InterruptedException {
                aWaits.incrementAndGet();
                super.await();
            }
        };
    }

    @AfterMethod
    public void cleanup() throws Exception {
        shiftEventsService.stop();
        ArtifactoryHome.unbind();
    }

    @Test
    public void testStop() throws Exception {
        // Given
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 1);

        // When
        shiftEventsService.stop();
        // Then
        Assert.assertTrue(shiftEventsService.executor.isTerminated());
    }

    @Test
    public void testHasPendingShifts() {
        // When
        shiftEventsService.triggerShift();
        // Then
        Assert.assertTrue(shiftEventsService.versionCache.hasNewVersion());
    }

    @Test
    public void testTriggerShiftAwakeThread() {
        // Given
        new Thread(() -> {
            try {
                Thread.sleep(1000);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
            shiftEventsService.triggerShift();
        }).start();
        Instant start = Instant.now();
        // When
        synchronized (shiftEventsService.lock) {
            try {
                shiftEventsService.lock.wait(5000);
            } catch (InterruptedException e) {
                Thread.currentThread().interrupt();
            }
        }
        // Then
        Assert.assertTrue(Duration.between(start, Instant.now()).toMillis() < 2000);
    }

    @Test
    public void testTriggerQueues() {

        Set<String> repos = Sets.newHashSet("repo1", "repo2");

        // Given
        Mockito.when(eventsService.shiftEvents()).thenAnswer(invocation -> {
            Set<String> r = new HashSet<>(repos);
            repos.clear();
            return r;
        });
        // When
        shiftEventsService.shiftEvents();
        // Then
        Mockito.verify(queueTriggeringService, Mockito.times(1))
                .triggerQueues(Mockito.eq(Sets.newHashSet("repo1", "repo2")));
    }

    @Test
    public void testStartIsTryingShift() throws Exception {

        // When
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() > 0);

        // Then
        Mockito.verify(shiftEventsService.eventsService, Mockito.times(1)).shiftEvents();
        Mockito.verify(shiftEventsService.queueTriggeringService, Mockito.never()).triggerQueues(Mockito.anySet());
    }

    @Test
    public void testTriggerShiftWakeupThreadAndShiftEvents() throws Exception {
        // Given
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() > 0);

        Thread.sleep(100);

        shiftEventsService.triggerShift();

        // Then
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 2);
        Mockito.verify(eventsService, Mockito.times(2)).shiftEvents();
        Mockito.verify(shiftEventsService.queueTriggeringService, Mockito.never()).triggerQueues(Mockito.anySet());
    }

    @Test
    public void testTriggerShiftDuringShiftInvokeAnotherShift() throws Exception {

        // Given
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 1);

        Thread.sleep(100);
        CountDownLatch count1 = new CountDownLatch(1);
        CountDownLatch count2 = new CountDownLatch(1);
        Mockito.doAnswer(invocation -> {
            count1.countDown();
            count2.await(5, TimeUnit.SECONDS);
            return Collections.emptySet();
        }).when(eventsService).shiftEvents();

        // When
        shiftEventsService.triggerShift();
        count1.await(5, TimeUnit.SECONDS);
        shiftEventsService.triggerShift();
        count2.countDown();

        // Then
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() >= 2);

        Mockito.verify(eventsService, Mockito.times(3)).shiftEvents();
        Mockito.verify(shiftEventsService.queueTriggeringService, Mockito.never()).triggerQueues(Mockito.anySet());
    }

    @Test
    public void testExceptionDuringShiftInvokeAnotherShift() throws Exception {

        // Given
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 1);

        Thread.sleep(100);
        AtomicInteger shifts = new AtomicInteger();

        Set<String> repos = Sets.newHashSet("repo1");

        Mockito.when(eventsService.shiftEvents()).thenAnswer((Answer<Set<String>>) invocation -> {
            if (shifts.incrementAndGet() == 1) {
                throw new RuntimeException("shift error");
            } else {
                Set<String> r = new HashSet<>(repos);
                repos.clear();
                return r;
            }
        });

        // When
        shiftEventsService.triggerShift();

        // Then
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 2);

        Mockito.verify(eventsService, Mockito.times(4)).shiftEvents();
        Mockito.verify(shiftEventsService.queueTriggeringService, Mockito.times(1))
                .triggerQueues(Mockito.anySet());
    }

    @Test
    public void testNoDistributedLockTryShiftAgain() throws Exception {
        // Given
        shiftEventsService.onContextCreated();
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() == 1);

        AtomicInteger locks = new AtomicInteger();
        Mockito.when(conflictGuard.tryToLock(Mockito.anyLong(), Mockito.any(TimeUnit.class))).thenAnswer(
                (Answer<Boolean>) invocation -> {
                    LoggerFactory.getLogger(ShiftEventsServiceImpl.class).debug("try lock ==> " + locks.get());
                    return locks.incrementAndGet() >= 3;
                });

        // When
        shiftEventsService.triggerShift();

        // Then
        Awaitility.await().atMost(5, TimeUnit.SECONDS).with()
                .pollInterval(1, TimeUnit.SECONDS)
                .until(() -> aWaits.get() >= 2);

        Mockito.verify(eventsService, Mockito.times(2)).shiftEvents();
        Mockito.verify(shiftEventsService.queueTriggeringService, Mockito.never()).triggerQueues(Mockito.anySet());
    }
}