package org.artifactory.metadata.service.provider;

import com.google.common.collect.ImmutableList;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.xray.XrayAddon;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.module.ModuleInfo;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoLayout;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.model.common.RepoPathImpl;
import org.artifactory.model.xstream.fs.FileInfoImpl;
import org.artifactory.model.xstream.fs.PropertiesImpl;
import org.artifactory.repo.Repo;
import org.artifactory.repo.db.DbLocalRepo;
import org.artifactory.repo.service.InternalRepositoryService;
import org.artifactory.spring.InternalArtifactoryContext;
import org.artifactory.storage.db.event.service.metadata.model.MutableMetadataEntityBOM;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.util.RepoLayoutUtils;
import org.jfrog.common.ClockUtils;
import org.mockito.Mockito;
import org.testng.annotations.Test;

import java.util.ArrayList;
import java.util.Collections;

import static org.artifactory.metadata.service.provider.DockerMetadataProvider.MANIFEST_TYPE_PROP;
import static org.mockito.Mockito.any;
import static org.mockito.Mockito.doNothing;
import static org.mockito.Mockito.doReturn;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;
import static org.testng.Assert.assertFalse;
import static org.testng.Assert.assertNull;
import static org.testng.Assert.assertTrue;

/**
 * @author Uriah Levy
 */
@Test
public class AbstractMetadataProviderTest extends ArtifactoryHomeBoundTest {

    public void testExcludedUserProperties() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("npm.description", "the gift is a cool movie");
        props.put("npm.keywords", "Gordo, Weirdo");
        props.put("npm.name", "Bojangles");

        npmMetadataProvider.userPropsResolver().apply(bom, props, null, null);

        assertEquals(bom.getUserProperties().size(), 0);

        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        bom = new MutableMetadataEntityBOM();
        props = new PropertiesImpl();
        props.put("rpm.metadata.name", "Robyn");
        props.put("rpm.metadata.summary", "the gift is a strange movie");

        rpmMetadataProvider.userPropsResolver().apply(bom, props, null, null);

        NugetMetadataProvider nugetMetadataProvider = new NugetMetadataProvider();
        bom = new MutableMetadataEntityBOM();
        props = new PropertiesImpl();
        props.put("nuget.dependency", "long, comma, separated, list, of, dependencies, to, exclude");

        nugetMetadataProvider.userPropsResolver().apply(bom, props, null, null);

        assertEquals(bom.getUserProperties().size(), 0);
    }

    public void testNullUserProperty() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("foo", null);
        props.put("foo", "bar");
        props.put("baz", "");
        props.put("boom", "box");

        npmMetadataProvider.userPropsResolver().apply(bom, props, null, null);

        assertEquals(bom.getUserProperties().size(), 2);
        assertEquals(bom.getUserProperties().get("foo").size(), 1);
        assertEquals(bom.getUserProperties().get("boom").size(), 1);
    }

    public void testQualifierResolver() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("npm.name", "@foo/bar");

        npmMetadataProvider.qualifiersResolver().apply(bom, props, null, null);

        assertEquals(1, bom.getQualifiers().size());
        assertEquals(bom.getQualifiers().get("npm_scope"), "foo");
    }

    public void testNpmTagsResolver() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("npm.keywords", "foo, bar");

        npmMetadataProvider.tagsResolver().apply(bom, props, null, null);

        assertEquals(2, bom.getTags().size());
        assertTrue(bom.getTags().containsAll(ImmutableList.of("foo", "bar")));

        props.put("npm.keywords", "\f\\oo//\\, /ba/r//");
        bom.getTags().clear();
        npmMetadataProvider.tagsResolver().apply(bom, props, null, null);

        assertEquals(2, bom.getTags().size());
        assertTrue(bom.getTags().containsAll(ImmutableList.of("foo", "bar")));

        props.clear();
        props.put("npm.keywords", "[]");
        bom.getTags().clear();
        npmMetadataProvider.tagsResolver().apply(bom, props, null, null);
        assertTrue(bom.getTags().isEmpty());

        props.clear();
        props.put("npm.keywords", "no_content");
        bom.getTags().clear();
        npmMetadataProvider.tagsResolver().apply(bom, props, null, null);
        assertTrue(bom.getTags().isEmpty());
    }

    public void testNugetTagsResolver() {
        NugetMetadataProvider nugetMetadataProvider = new NugetMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("nuget.tags", "foo bar");

        nugetMetadataProvider.tagsResolver().apply(bom, props, null, null);

        assertEquals(2, bom.getTags().size());
        assertTrue(bom.getTags().containsAll(ImmutableList.of("foo", "bar")));
    }

    public void descriptionResolver() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        PropertiesImpl props = new PropertiesImpl();
        props.put("npm.description", "darn cool package");

        npmMetadataProvider.packageDescriptionResolver().apply(bom, props, null, null);

        assertEquals(bom.getDescription(), "darn cool package");

        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        bom = new MutableMetadataEntityBOM();
        props = new PropertiesImpl();
        props.put("rpm.metadata.name", "Robyn");
        props.put("rpm.metadata.summary", "the gift is a strange movie");

        rpmMetadataProvider.packageDescriptionResolver().apply(bom, props, null, null);

        assertEquals(bom.getDescription(), "the gift is a strange movie");

        props.clear();
        bom.setDescription(null);
        npmMetadataProvider.packageDescriptionResolver().apply(bom, props, null, null);

        assertNull(bom.getDescription());

        //new:
        props.clear();
        props.put("npm.description", "no_content");
        npmMetadataProvider.packageDescriptionResolver().apply(bom, props, null, null);
        assertNull(bom.getDescription());
    }

    public void testNpmPkgIdResolver() {
        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("npm-local", "express/-/express-1.0.0.tgz"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("npm.name", "express");

        npmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "npm://express");

        props.replaceValues("npm.name", ImmutableList.of(new String(new byte[500])));
        npmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertTrue(bom.getPkgid().startsWith("npm://"));
        assertEquals(bom.getPkgid().length(), 255);
    }

    public void testDebianFileInfoResolver() {
        ArtifactoryContext artifactoryContext = Mockito.mock(InternalArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        RepositoryService repositoryService = mock(RepositoryService.class);
        LocalRepoDescriptor localRepoDescriptor = mock(LocalRepoDescriptor.class);
        when(localRepoDescriptor.isLocal()).thenReturn(true);
        when(localRepoDescriptor.isCache()).thenReturn(true);
        when(artifactoryContext.beanForType(RepositoryService.class)).thenReturn(repositoryService);
        when(repositoryService.localOrCachedRepoDescriptorByKey(any())).thenReturn(localRepoDescriptor);

        DebianMetadataProvider debianMetadataProvider = new DebianMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("debian-local", "lucid/i386/acl-1.0.0.deb"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("deb.distribution", "lucid");
        props.put("deb.architecture", "i386");

        debianMetadataProvider.fileInfoResolver().apply(bom, props, fileInfo, null);
        debianMetadataProvider.architectureResolver().apply(bom, props, fileInfo, null);
        debianMetadataProvider.distributionResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getArtifactMetadata().get(0).getDist(),"lucid");
        assertEquals(bom.getArtifactMetadata().get(0).getArch(),"i386");
    }

    public void testDebianPkgIdResolver() {
        DebianMetadataProvider debianMetadataProvider = new DebianMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("debian-local", "lucid/i386/acl-1.0.0.deb"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("deb.name", "acl");

        debianMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "deb://acl");
    }

    public void testAlpineFileInfoResolver() {
        ArtifactoryContext artifactoryContext = Mockito.mock(InternalArtifactoryContext.class);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
        RepositoryService repositoryService = mock(RepositoryService.class);
        LocalRepoDescriptor localRepoDescriptor = mock(LocalRepoDescriptor.class);
        when(localRepoDescriptor.isLocal()).thenReturn(true);
        when(localRepoDescriptor.isCache()).thenReturn(true);
        when(artifactoryContext.beanForType(RepositoryService.class)).thenReturn(repositoryService);
        when(repositoryService.localOrCachedRepoDescriptorByKey(any())).thenReturn(localRepoDescriptor);

        AlpineMetadataProvider alpineMetadataProvider = new AlpineMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo =
                new FileInfoImpl(new RepoPathImpl("alpine-local", "v3.11/main/x86_64/unzip-6.0-r6.apk"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("alpine.branch", "v3.11");
        props.put("alpine.architecture", "x86_64");

        alpineMetadataProvider.fileInfoResolver().apply(bom, props, fileInfo, null);
        alpineMetadataProvider.distributionResolver().apply(bom, props, fileInfo, null);
        alpineMetadataProvider.architectureResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getArtifactMetadata().get(0).getDist(), "v3.11");
        assertEquals(bom.getArtifactMetadata().get(0).getArch(), "x86_64");
    }

    public void testAlpinePkgIdResolver() {
        AlpineMetadataProvider alpineMetadataProvider = new AlpineMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo =
                new FileInfoImpl(new RepoPathImpl("alpine-local", "v3.11/main/x86_64/unzip-6.0-r6.apk"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("alpine.name", "unzip");

        alpineMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "alpine://unzip");
    }

    public void testHelmPkgIdResolver() {
        HelmMetadataProvider helmMetadataProvider = new HelmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("helm-local", "artifactory-helm/artifactory-7.0.0.tgz"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("chart.name", "artifactory");

        helmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "helm://artifactory");
    }

    public void testComposerPkgIdResolver() {
        ComposerMetadataProvider composerMetadataProvider = new ComposerMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("composer-local", "jfi/jfi-core/jfi-code-2.1.250.tar.gz"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("composer.name", "jfrog/insight-core");

        composerMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "composer://jfrog/insight-core");
    }

    public void testOpkgPkgIdResolver() {
        OpkgMetadataProvider opkgMetadataProvider = new OpkgMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("opkg-local", "test-package1_0.1_x86.ipk"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("opkg.name", "test-package1");

        opkgMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "opkg://test-package1");
    }

    public void testPuppetPkgIdResolver() {
        PuppetMetadataProvider puppetMetadataProvider = new PuppetMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("puppet-local", "puppetlabs-mysql-3.9.0.tar.gz"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("puppet.name", "puppetlabs-mysql");

        puppetMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "puppet://puppetlabs-mysql");
    }

    public void testVagrantPkgIdResolver() {
        VagrantMetadataProvider vagrantMetadataProvider = new VagrantMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("vagrant-local", "trusty64_1.0.0_virtualbox.box"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("box_name", "trusty");

        vagrantMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "vagrant://trusty");
    }

    public void testIvyPkgIdResolver() {
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        IvyMetadataProvider ivyMetadataProvider = new IvyMetadataProvider(internalRepositoryService);
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("ivy-local", "org.apache.ivy.example/version/1.0-dev-b1/ivys/ivy-0.1-deb-b1.xml"));

        Repo repo = mock(Repo.class);
        when(internalRepositoryService.repositoryByKey("ivy-local")).thenReturn(repo);
        when(repo.getDescriptor()).thenReturn(new LocalRepoDescriptor());
        when(repo.getItemModuleInfo("org.apache.ivy.example/version/1.0-dev-b1/ivys/ivy-0.1-deb-b1.xml"))
                .thenReturn(new ModuleInfo("org.apache.ivy.example", "version", "1.0-dev-b1", null, null, null, "xml",
                        "xml",
                        Collections.emptyMap()));

        ivyMetadataProvider.nameResolver().apply(bom, null, fileInfo, null);
        ivyMetadataProvider.versionResolver().apply(bom, null, fileInfo, null);
        ivyMetadataProvider.pkgIdResolver().apply(bom, null, fileInfo, null);

        assertEquals(bom.getName(), "version");
        assertEquals(bom.getVersion(), "1.0-dev-b1");
        assertEquals(bom.getPkgid(), "ivy://org.apache.ivy.example:version");
    }

    public void testSbtPkgIdResolver() {
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        SbtMetadataProvider sbtMetadataProvider = new SbtMetadataProvider(internalRepositoryService);
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("sbt-local",
                        "sbt_example/sbt_example_2.11/1.1-SNAPSHOT/sbt_example_2.11-1.1-SNAPSHOT.pom"));

        Repo repo = mock(Repo.class);
        when(internalRepositoryService.repositoryByKey("sbt-local")).thenReturn(repo);
        when(repo.getDescriptor()).thenReturn(new LocalRepoDescriptor());
        when(repo.getItemModuleInfo("sbt_example/sbt_example_2.11/1.1-SNAPSHOT/sbt_example_2.11-1.1-SNAPSHOT.pom"))
                .thenReturn(new ModuleInfo("sbt_example", "sbt_example_2.11", "1.1-SNAPSHOT", null, null, null, "pom",
                        "pom",
                        Collections.emptyMap()));

        sbtMetadataProvider.nameResolver().apply(bom, null, fileInfo, null);
        sbtMetadataProvider.versionResolver().apply(bom, null, fileInfo, null);
        sbtMetadataProvider.pkgIdResolver().apply(bom, null, fileInfo, null);

        assertEquals(bom.getName(), "sbt_example_2.11");
        assertEquals(bom.getVersion(), "1.1-SNAPSHOT");
        assertEquals(bom.getPkgid(), "gav://sbt_example:sbt_example_2.11");
    }

    public void testGradlePkgIdResolver() {
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        GradleMetadataProvider gradleMetadataProvider = new GradleMetadataProvider(internalRepositoryService);
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("gradle-local",
                        "gradle_example/gradle_example_2.11/1.1-SNAPSHOT/gradle_example_2.11-1.1-SNAPSHOT.pom"));

        Repo repo = mock(Repo.class);
        when(internalRepositoryService.repositoryByKey("gradle-local")).thenReturn(repo);
        when(repo.getDescriptor()).thenReturn(new LocalRepoDescriptor());
        when(repo.getItemModuleInfo(
                "gradle_example/gradle_example_2.11/1.1-SNAPSHOT/gradle_example_2.11-1.1-SNAPSHOT.pom"))
                .thenReturn(
                        new ModuleInfo("gradle_example", "gradle_example_2.11", "1.1-SNAPSHOT", null, null, null, "pom",
                                "pom",
                                Collections.emptyMap()));

        gradleMetadataProvider.nameResolver().apply(bom, null, fileInfo, null);
        gradleMetadataProvider.versionResolver().apply(bom, null, fileInfo, null);
        gradleMetadataProvider.pkgIdResolver().apply(bom, null, fileInfo, null);

        assertEquals(bom.getName(), "gradle_example_2.11");
        assertEquals(bom.getVersion(), "1.1-SNAPSHOT");
        assertEquals(bom.getPkgid(), "gav://gradle_example:gradle_example_2.11");
    }

    public void testConanPkgIdResolver() {
        ConanMetadataProvider conanMetadataProvider = new ConanMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("conan-local", "conan/zlib/1.2.9/stable/0/export/conanfile.py"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("conan.package.name", "zlib");

        conanMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "conan://zlib");
    }

    public void testConanPkgIdWithVendorResolver() {
        ConanMetadataProvider conanMetadataProvider = new ConanMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("conan-local", "conan/zlib/1.2.9/stable/0/export/conanfile.py"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("conan.package.name", "zlib");
        props.put("conan.package.vendor", "jfrog");

        conanMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "conan://jfrog:zlib");
    }

    public void testGoPkgIdResolver() {
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        GoMetadataProvider goMetadataProvider = new GoMetadataProvider(internalRepositoryService);
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("go-local", "jfrog.com/jfrog-go-commons/@v/v0.0.1-20190617090946-29.mod"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("go.name", "jfrog.com/jfrog-go-commons");

        goMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);

        assertEquals(bom.getPkgid(), "go://jfrog.com/jfrog-go-commons");
    }

    public void testTrimValueIfNeeded() {
        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        String normalizedValue = rpmMetadataProvider
                .trimValueIfNeeded("rpm.metadata.summary", new String(new byte[5000]));
        assertEquals(normalizedValue.length(), 5000);

        normalizedValue = rpmMetadataProvider
                .trimValueIfNeeded("rpm.metadata.tooLong", new String(new byte[5000]));
        assertEquals(normalizedValue.length(), 4000);

        normalizedValue = rpmMetadataProvider
                .trimValueIfNeeded("rpm.metadata.notTooLong", new String(new byte[5]));
        assertEquals(normalizedValue.length(), 5);

        DockerMetadataProvider dockerMetadataProvider = new DockerMetadataProvider();
        normalizedValue = dockerMetadataProvider
                .trimValueIfNeeded("docker.hasNoDescriptionKey", new String(new byte[5000]));
        assertEquals(normalizedValue.length(), 4000);

        normalizedValue = dockerMetadataProvider
                .trimValueIfNeeded("docker.hasNoDescriptionKey", new String(new byte[4]));
        assertEquals(normalizedValue.length(), 4);

        NpmMetadataProvider npmMetadataProvider = new NpmMetadataProvider();
        normalizedValue = npmMetadataProvider.trimValueIfNeeded("npm.keywords", new String(new byte[5000]));
        assertEquals(normalizedValue.length(), 5000);

        normalizedValue = npmMetadataProvider.trimValueIfNeeded("npm.description", new String(new byte[5000]));
        assertEquals(normalizedValue.length(), 5000);
    }

    public void testRpmPkgIdAndNameResolver() {
        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("rpm-local", "ImageMagick.rpm"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("rpm.metadata.name", "ImageMagick");
        props.put("rpm.metadata.release", "2.el7");

        rpmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getPkgid(), "rpm://7:ImageMagick");

        rpmMetadataProvider.nameResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getName(), "7:ImageMagick");

        props = new PropertiesImpl();
        props.put("rpm.metadata.name", "ImageMagick");
        props.put("rpm.metadata.release", "2.el7.AnotherValidElVersion");

        rpmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getPkgid(), "rpm://7:ImageMagick");

        props = new PropertiesImpl();
        props.put("rpm.metadata.name", "ImageMagick");

        rpmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);
        rpmMetadataProvider.nameResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getPkgid(), "rpm://ImageMagick");
        assertEquals(bom.getName(), "ImageMagick");

        props = new PropertiesImpl();
        props.put("rpm.metadata.name", "dejavu-font");
        props.put("rpm.metadata.release", "9.el7.centos");

        rpmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getPkgid(), "rpm://7:dejavu-font");
    }

    public void testGradleOnlySupportsMavenRepo() {
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        GradleMetadataProvider gradleMetadataProvider = new GradleMetadataProvider(internalRepositoryService);

        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.Gradle);
        localRepoDescriptor.setRepoLayout(RepoLayoutUtils.MAVEN_2_DEFAULT);
        when(internalRepositoryService.repoDescriptorByKey("gradle-local")).thenReturn(localRepoDescriptor);

        assertTrue(gradleMetadataProvider.isLeadArtifact(new RepoPathImpl("gradle-local", "package.pom")));
        assertFalse(gradleMetadataProvider.isLeadArtifact(new RepoPathImpl("gradle-local", "package.not_a_pom")));

        localRepoDescriptor.setRepoLayout(RepoLayoutUtils.GRADLE_DEFAULT_DEPRECATED);
        assertFalse(gradleMetadataProvider.isLeadArtifact(new RepoPathImpl("gradle-local", "package.pom")));
    }

    @SuppressWarnings("unchecked")
    public void testLayoutBasedProvidersLayoutRestricted() {
        InternalRepositoryService repoService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);
        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);

        // Ivy: maven layout
        IvyMetadataProvider ivyMetadataProvider = new IvyMetadataProvider(repoService);
        LocalRepoDescriptor localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.Ivy);
        localRepoDescriptor.setRepoLayout(RepoLayoutUtils.IVY_DEFAULT);
        when(repoService.repoDescriptorByKey("ivy-local")).thenReturn(localRepoDescriptor);

        RepoPathImpl leadArtifactRepoPath = new RepoPathImpl("ivy-local", "org/jfrog/package/1.0/package-1.0.pom");
        assertTrue(
                ivyMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());

        // Ivy: some other layout
        RepoLayout unsupportedLayout = new RepoLayout();
        unsupportedLayout.setName("unsupported-layout");
        localRepoDescriptor.setRepoLayout(unsupportedLayout);

        assertFalse(
                ivyMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath)
                        .isPresent());

        // Ivy: only depth > 2 is supported for the lead artifact
        assertFalse(ivyMetadataProvider.buildVersionFilesQuery(new RepoPathImpl("ivy-local", "flat-path.pom"))
                .isPresent());

        // Maven: maven layout
        MavenMetadataProvider mavenMetadataProvider = new MavenMetadataProvider(repoService, addonsManagerMock);
        localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.Maven);
        localRepoDescriptor.setRepoLayout(RepoLayoutUtils.MAVEN_2_DEFAULT);
        when(repoService.repoDescriptorByKey("maven-local")).thenReturn(localRepoDescriptor);
        leadArtifactRepoPath = new RepoPathImpl("maven-local", "com/jfrog/package/1.0/package-1.0.pom");
        DbLocalRepo<LocalRepoDescriptor> repo = mock(DbLocalRepo.class);
        when(repo.getDescriptor()).thenReturn(localRepoDescriptor);
        when(repo.getItemModuleInfo("com/jfrog/package/1.0/package-1.0.pom"))
                .thenReturn(new ModuleInfo("com/jfrog", "artifactory", "1.0", null, null, null, "pom", "pom",
                        Collections.emptyMap()));
        when(repoService.repositoryByKey(leadArtifactRepoPath.getRepoKey())).thenReturn(repo);

        assertTrue(
                mavenMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());

        // Maven: some other layout
        localRepoDescriptor.setRepoLayout(RepoLayoutUtils.IVY_DEFAULT);

        assertFalse(
                mavenMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());

        // Go: Go layout
        GoMetadataProvider goMetadataProvider = new GoMetadataProvider(repoService);
        localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.Go);
        RepoLayout goLayout = new RepoLayout();
        goLayout.setName(RepoLayoutUtils.GO_DEFAULT_NAME);
        localRepoDescriptor.setRepoLayout(goLayout);
        when(repoService.repoDescriptorByKey("go-local")).thenReturn(localRepoDescriptor);
        repo = mock(DbLocalRepo.class);
        leadArtifactRepoPath =
                new RepoPathImpl("go-local", "jfrog.com/jfrog-go-commons/@v/v0.0.1-20190617090946-29.mod");
        when(repo.getDescriptor()).thenReturn(localRepoDescriptor);
        when(repo.getItemModuleInfo("jfrog.com/jfrog-go-commons/@v/v0.0.1-20190617090946-29.mod"))
                .thenReturn(
                        new ModuleInfo("jfrog.com", "jfrog-go-commons", "0.0.1", null, "20190617090946-29", null, "mod",
                                "mod",
                                Collections.emptyMap()));
        when(repoService.repositoryByKey(leadArtifactRepoPath.getRepoKey())).thenReturn(repo);

        assertTrue(
                goMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());

        // Go: some other layout
        localRepoDescriptor.setRepoLayout(unsupportedLayout);
        assertFalse(
                goMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());

        // SBT: SBT layout
        SbtMetadataProvider sbtMetadataProvider = new SbtMetadataProvider(repoService);
        localRepoDescriptor = new LocalRepoDescriptor();
        localRepoDescriptor.setType(RepoType.SBT);
        RepoLayout sbtLayout = new RepoLayout();
        sbtLayout.setName(RepoLayoutUtils.SBT_DEFAULT_NAME);
        localRepoDescriptor.setRepoLayout(sbtLayout);
        when(repoService.repoDescriptorByKey("sbt-local")).thenReturn(localRepoDescriptor);
        leadArtifactRepoPath = new RepoPathImpl("sbt-local",
                "sbt_example/sbt_example_2.11/1.1-SNAPSHOT/sbt_example_2.11-1.1-SNAPSHOT.pom");
        repo = mock(DbLocalRepo.class);
        when(repo.getDescriptor()).thenReturn(localRepoDescriptor);
        when(repo.getItemModuleInfo("sbt_example/sbt_example_2.11/1.1-SNAPSHOT/sbt_example_2.11-1.1-SNAPSHOT.pom"))
                .thenReturn(
                        new ModuleInfo("sbt_example", "sbt_example_2.11", "1.1", "SNAPSHOT", "SNAPSHOT", null, "pom",
                                "pom",
                                Collections.emptyMap()));
        when(repoService.repositoryByKey(leadArtifactRepoPath.getRepoKey())).thenReturn(repo);

        assertTrue(
                sbtMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());
        // SBT: some other layout
        localRepoDescriptor.setRepoLayout(unsupportedLayout);
        assertFalse(
                sbtMetadataProvider.buildVersionFilesQuery(leadArtifactRepoPath).isPresent());
    }

    public void testRpmPkgIdWithComplexResolver() {
        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("rpm-local", "ImageMagick.rpm"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("rpm.metadata.name", "ImageMagick");
        props.put("rpm.metadata.release", "2.cmplx1234.el6");
        rpmMetadataProvider.pkgIdResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getPkgid(), "rpm://6:ImageMagick");
    }

    public void testRpmVersionResolver() {
        RpmMetadataProvider rpmMetadataProvider = new RpmMetadataProvider();
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl("rpm-local", "ImageMagick.rpm"));
        PropertiesImpl props = new PropertiesImpl();
        props.put("rpm.metadata.name", "ImageMagick");
        props.put("rpm.metadata.version", "1.5.3");
        props.put("rpm.metadata.epoch", "1");
        props.put("rpm.metadata.release", "2.el7");

        rpmMetadataProvider.versionResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getVersion(), "1:1.5.3-2.el7");

        props = new PropertiesImpl();
        bom = new MutableMetadataEntityBOM();
        rpmMetadataProvider.versionResolver().apply(new MutableMetadataEntityBOM(), props, fileInfo, null);
        assertNull(bom.getVersion());

        props = new PropertiesImpl();
        bom = new MutableMetadataEntityBOM();
        props.put("rpm.metadata.version", "1.5.3");
        props.put("rpm.metadata.epoch", "0");
        rpmMetadataProvider.versionResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getVersion(), "0:1.5.3");

        props = new PropertiesImpl();
        bom = new MutableMetadataEntityBOM();
        props.put("rpm.metadata.version", "1.5.3");
        rpmMetadataProvider.versionResolver().apply(bom, props, fileInfo, null);
        assertEquals(bom.getVersion(), "1.5.3");
    }

    public void testMavenVersionPopulationUniqueSnapshotSupported() {
        Repo repo = mock(Repo.class);
        ModuleInfo moduleInfo = new ModuleInfo(
                "org.jfrog.test",
                "multi3",
                "2.17",
                "SNAPSHOT",
                "20200402.123134-8", null, "pom", null, Collections.emptyMap());
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey(any())).thenReturn(repo);
        when(xrayAddonMock.isXrayEnabled()).thenReturn(true);
        when(xrayAddonMock.isXrayOfMinimalRequiredVersion(any())).thenReturn(true);
        when(repo.getItemModuleInfo(any())).thenReturn(moduleInfo);

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl(
                "maven-local",
                "org/jfrog/test/multi3/2.17-SNAPSHOT/multi3-2.17-20200402.123134-8.pom"
        ));
        provider.versionResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getVersion(), "2.17-20200402.123134-8");
    }

    public void testMavenVersionPopulationUniqueSnapshotSupportedNoXrayConfigured() {
        Repo repo = mock(Repo.class);
        ModuleInfo moduleInfo = new ModuleInfo(
                "org.jfrog.test",
                "multi3",
                "2.17",
                "SNAPSHOT",
                "20200402.123134-8", null, "pom", null, Collections.emptyMap());
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey(any())).thenReturn(repo);
        when(xrayAddonMock.isDefault()).thenReturn(false);
        when(xrayAddonMock.isXrayConfigMissing()).thenReturn(true);
        when(repo.getItemModuleInfo(any())).thenReturn(moduleInfo);

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl(
                "maven-local",
                "org/jfrog/test/multi3/2.17-SNAPSHOT/multi3-2.17-20200402.123134-8.pom"
        ));
        provider.versionResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getVersion(), "2.17-20200402.123134-8");
    }

    public void testMavenVersionPopulationUniqueSnapshotSupportedOSS() {
        Repo repo = mock(Repo.class);
        ModuleInfo moduleInfo = new ModuleInfo(
                "org.jfrog.test",
                "multi3",
                "2.17",
                "SNAPSHOT",
                "20200402.123134-8", null, "pom", null, Collections.emptyMap());
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey(any())).thenReturn(repo);
        when(xrayAddonMock.isDefault()).thenReturn(true);
        when(repo.getItemModuleInfo(any())).thenReturn(moduleInfo);

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl(
                "maven-local",
                "org/jfrog/test/multi3/2.17-SNAPSHOT/multi3-2.17-20200402.123134-8.pom"
        ));
        provider.versionResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getVersion(), "2.17-20200402.123134-8");
    }

    public void testMavenVersionPopulationUniqueSnapshotSupportedXrayDisabled() {
        Repo repo = mock(Repo.class);
        ModuleInfo moduleInfo = new ModuleInfo(
                "org.jfrog.test",
                "multi3",
                "2.17",
                "SNAPSHOT",
                "20200402.123134-8", null, "pom", null, Collections.emptyMap());
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey(any())).thenReturn(repo);
        when(xrayAddonMock.isDefault()).thenReturn(false);
        when(xrayAddonMock.isXrayConfigMissing()).thenReturn(false);
        when(xrayAddonMock.isXrayEnabled()).thenReturn(false);
        when(repo.getItemModuleInfo(any())).thenReturn(moduleInfo);

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl(
                "maven-local",
                "org/jfrog/test/multi3/2.17-SNAPSHOT/multi3-2.17-20200402.123134-8.pom"
        ));
        provider.versionResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getVersion(), "2.17-20200402.123134-8");
    }

    public void testMavenVersionPopulationUniqueSnapshotUnsupportedXrayLowerThanMinVersion() {
        Repo repo = mock(Repo.class);
        ModuleInfo moduleInfo = new ModuleInfo(
                "org.jfrog.test",
                "multi3",
                "2.17",
                "SNAPSHOT",
                "20200402.123134-8", null, "pom", null, Collections.emptyMap());
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey(any())).thenReturn(repo);
        when(xrayAddonMock.isXrayEnabled()).thenReturn(true);
        when(xrayAddonMock.isXrayOfMinimalRequiredVersion(any())).thenReturn(false);
        when(repo.getItemModuleInfo(any())).thenReturn(moduleInfo);

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(new RepoPathImpl(
                "maven-local",
                "org/jfrog/test/multi3/2.17-SNAPSHOT/multi3-2.17-20200402.123134-8.pom"
        ));
        provider.versionResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getVersion(), "2.17-SNAPSHOT");
    }

    public void testLayoutBasedResolvers() {
        Repo repo = mock(Repo.class);
        InternalRepositoryService internalRepositoryService = mock(InternalRepositoryService.class);
        XrayAddon xrayAddonMock = mock(XrayAddon.class);
        AddonsManager addonsManagerMock = mock(AddonsManager.class);

        when(addonsManagerMock.addonByType(XrayAddon.class)).thenReturn(xrayAddonMock);
        when(internalRepositoryService.repositoryByKey("libs-release-local")).thenReturn(repo);
        when(repo.getDescriptor()).thenReturn(new LocalRepoDescriptor());
        when(repo.getItemModuleInfo("org/jfrog/artifactory/1.0.0/artifactory-1.0.0.pom"))
                .thenReturn(new ModuleInfo(
                        "org/jfrog",
                        "artifactory",
                        "1.0.0",
                        null,
                        null,
                        null,
                        "pom",
                        "pom",
                        Collections.emptyMap()));

        MavenMetadataProvider provider = new MavenMetadataProvider(internalRepositoryService, addonsManagerMock);

        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("libs-release-local", "org/jfrog/artifactory/1.0.0/artifactory-1.0.0.pom"));

        provider.nameResolver().apply(bom, null, fileInfo, null);
        provider.versionResolver().apply(bom, null, fileInfo, null);
        provider.pkgIdResolver().apply(bom, null, fileInfo, null);

        assertEquals(bom.getName(), "artifactory");
        assertEquals(bom.getVersion(), "1.0.0");
        assertEquals(bom.getPkgid(), "gav://org.jfrog:artifactory");
    }

    public void testLastModifiedMappedToVersionModified() {
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("libs-release-local", "org/jfrog/artifactory/1.0.0/artifactory-1.0.0.pom"));
        long lastModified = ClockUtils.epochMillis();
        fileInfo.setLastModified(lastModified);
        DockerMetadataProvider dockerMetadataProvider = spy(new DockerMetadataProvider());

        doReturn(new ArrayList<>()).when(dockerMetadataProvider).findVersionFiles(any());
        doNothing().when(dockerMetadataProvider).setVersionRepo(any(), any(), any());

        dockerMetadataProvider.fileInfoResolver().apply(bom, new PropertiesImpl(), fileInfo, null);
        assertEquals(bom.getModified(), lastModified);
    }

    public void testColonReplacedWithDoubleUnderscoreForFatManifest() {
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("docker-local",
                        "ubuntu/sha256__2f6fb811f836424acf6120ee35e3698ee65861f290332cb009c709c3d5088a82/manifest" +
                                ".json"));
        DockerMetadataProvider dockerMetadataProvider = spy(new DockerMetadataProvider());

        doReturn(new ArrayList<>()).when(dockerMetadataProvider).findVersionFiles(any());
        doNothing().when(dockerMetadataProvider).setVersionRepo(any(), any(), any());

        PropertiesImpl properties = new PropertiesImpl();
        properties.put(dockerMetadataProvider.getType(properties) +
                        DockerMetadataProvider.DockerTypeSpecificMetadataTemplate.VERSION_SUFFIX,
                "sha256:2f6fb811f836424acf6120ee35e3698ee65861f290332cb009c709c3d5088a82");
        dockerMetadataProvider.versionResolver().apply(bom, properties, fileInfo, null);
        assertEquals(bom.getVersion(), "sha256__2f6fb811f836424acf6120ee35e3698ee65861f290332cb009c709c3d5088a82");
    }

    public void testVersionUntouchedForNonFatManifest() {
        MutableMetadataEntityBOM bom = new MutableMetadataEntityBOM();
        FileInfoImpl fileInfo = new FileInfoImpl(
                new RepoPathImpl("docker-local",
                        "ubuntu/latest/manifest.json"));
        DockerMetadataProvider dockerMetadataProvider = spy(new DockerMetadataProvider());

        doReturn(new ArrayList<>()).when(dockerMetadataProvider).findVersionFiles(any());
        doNothing().when(dockerMetadataProvider).setVersionRepo(any(), any(), any());

        PropertiesImpl properties = new PropertiesImpl();
        properties.put(dockerMetadataProvider.getType(properties) +
                        DockerMetadataProvider.DockerTypeSpecificMetadataTemplate.VERSION_SUFFIX,
                "latest");
        dockerMetadataProvider.versionResolver().apply(bom, properties, fileInfo, null);
        assertEquals(bom.getVersion(), "latest");
    }

    public void testDockerGetType() {
        DockerMetadataProvider dockerMetadataProvider = new DockerMetadataProvider();
        PropertiesImpl properties = new PropertiesImpl();

        properties.put(MANIFEST_TYPE_PROP, "application/vnd.oci.image.manifest.v1+json");
        assertEquals(dockerMetadataProvider.getType(properties), "oci");

        properties.replaceValues(MANIFEST_TYPE_PROP, ImmutableList.of("application/vnd.oci.image.index.v1+json"));
        assertEquals(dockerMetadataProvider.getType(properties), "oci");

        properties.replaceValues(MANIFEST_TYPE_PROP,
                ImmutableList.of("application/vnd.docker.distribution.manifest.v2+json"));
        assertEquals(dockerMetadataProvider.getType(properties), "docker");

        properties.replaceValues(MANIFEST_TYPE_PROP,
                ImmutableList.of("application/vnd.docker.distribution.manifest.list.v2+json"));
        assertEquals(dockerMetadataProvider.getType(properties), "docker");

        properties.replaceValues(MANIFEST_TYPE_PROP, ImmutableList.of("some value"));
        assertEquals(dockerMetadataProvider.getType(properties), "docker");

        properties.removeAll(MANIFEST_TYPE_PROP);
        assertEquals(dockerMetadataProvider.getType(properties), "docker");
    }

    public void testIsLeadArtifact() {
        PypiMetadataProvider pypiMetadataProvider = new PypiMetadataProvider();
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.whl")));
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.tgz")));
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.tar.gz")));
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.egg")));
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.bz2")));
        assertTrue(pypiMetadataProvider.isLeadArtifact(new RepoPathImpl("pypi-local", "package.gz")));

        ChefMetadataProvider chefMetadataProvider = new ChefMetadataProvider();
        assertTrue(chefMetadataProvider.isLeadArtifact(new RepoPathImpl("chef-local", "package.tgz")));
        assertTrue(chefMetadataProvider.isLeadArtifact(new RepoPathImpl("chef-local", "package.tar.gz")));
        assertFalse(chefMetadataProvider.isLeadArtifact(new RepoPathImpl("chef-local", "package.pom")));

        BowerMetadataProvider bowerMetadataProvider = new BowerMetadataProvider();
        assertTrue(bowerMetadataProvider.isLeadArtifact(new RepoPathImpl("bower-local", "package.tgz")));
        assertTrue(bowerMetadataProvider.isLeadArtifact(new RepoPathImpl("bower-local", "package.tar.gz")));

        CocoaPodsMetadataProvider cocoaPodsMetadataProvider = new CocoaPodsMetadataProvider();
        assertTrue(cocoaPodsMetadataProvider.isLeadArtifact(new RepoPathImpl("cocoa-local", "package.tgz")));
        assertTrue(cocoaPodsMetadataProvider.isLeadArtifact(new RepoPathImpl("cocoa-local", "package.tar.gz")));

        DockerMetadataProvider dockerMetadataProvider = new DockerMetadataProvider();
        assertTrue(dockerMetadataProvider.isLeadArtifact(new RepoPathImpl("docker-local", "manifest.json")));
        assertFalse(dockerMetadataProvider.isLeadArtifact(new RepoPathImpl("docker-local", "no_manifest.json")));

        IvyMetadataProvider ivyMetadataProvider = new IvyMetadataProvider(null);
        assertTrue(ivyMetadataProvider.isLeadArtifact(new RepoPathImpl("ivy-local", "ivy-something.xml")));
        assertTrue(ivyMetadataProvider.isLeadArtifact(new RepoPathImpl("ivy-local", "ivy.xml")));
    }

    public void testIsLeadNoExtension() {
        BowerMetadataProvider bowerMetadataProvider = new BowerMetadataProvider();
        assertFalse(bowerMetadataProvider.isLeadArtifact(new RepoPathImpl("bower-local", "package")));
    }
}
