package org.artifactory.webapp.main;
/*
 *
 * Copyright 2018 JFrog Ltd. All rights reserved.
 * JFROG PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */

import org.apache.commons.lang3.SystemUtils;
import org.artifactory.webapp.WebappUtils;
import org.jfrog.common.ExecutionUtils;
import org.jfrog.common.FileUtils;
import org.jfrog.common.RetryException;
import org.jfrog.common.RouterUtils;
import org.jfrog.common.platform.test.helpers.ProcessUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zeroturnaround.exec.ProcessExecutor;
import org.zeroturnaround.exec.StartedProcess;

import java.io.File;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.CompletableFuture;

/**
 * @author Gal Ben Ami
 */
public class ReplicatorProcess {

    private final static Logger log = LoggerFactory.getLogger(ReplicatorProcess.class);


    private StartedProcess process;
    private File jfProductHomeDir;
    private File replicatorBinDir;
    private File binFile;
    private File replicatorEtcDir;


    public ReplicatorProcess(File jfProductHome) {
        this.jfProductHomeDir = jfProductHome;
    }


    public void setupAndStartReplicator() {
        createHome();
        prepareBinary();
        writeJoinKeyToFile();
        startReplicator();
    }

    private void startReplicator() {
        try {
            CompletableFuture<Boolean> future = ExecutionUtils.retry(() -> {
                executeStartReplicator();
                return null;
            }, ExecutionUtils.RetryOptions.builder().numberOfRetries(3).timeout(1000).build());

            future.get();
        } catch (Exception e) {
            throw new RuntimeException("Could not start replicator", e);
        }
    }

    private void executeStartReplicator() throws RetryException {
        List<String> cmd = new ArrayList<>();
        cmd.add(binFile.getAbsolutePath());
        try {
            log.info(cmd.toString());
            ProcessExecutor processConfig = new ProcessExecutor()
                    .command(cmd)
                    .environment("JF_PRODUCT_HOME", jfProductHomeDir.getPath())
                    .environment("JF_REPLICATOR_PORT", String.valueOf(9001))
                    .environment("JF_REPLICATOR_LOGGING_APPLICATION_LEVEL", "debug")
                    .environment("JF_REPLICATOR_P2P_TRACKER_ENABLED", "true")
                    .environment("JF_REPLICATOR_P2P_TRACKER_JOINKEY", "ed7069465f34a2fc6ab98423485f38b2")
                    .redirectOutput(System.out)
                    .redirectError(System.err)
                    .destroyOnExit();

            processConfig.environment(RouterUtils.JFROG_ROUTER_URL, "localhost:8080");

            this.process = processConfig.start();

        } catch (Exception e) {
            throw new RetryException("Could not start replicator", e);
        }
    }

    private void writeJoinKeyToFile() {
        ProcessUtils.writeTestJoinKeyToFile(Paths.get(jfProductHomeDir.getPath(), "var").toFile(),
                AccessProcess.TEST_JFROG_JOIN_KEY);
    }

    private void prepareBinary() {
        String absolutePath = WebappUtils.getReplicatorApiExecutable(SystemUtils.IS_OS_MAC).getAbsolutePath();
        binFile = new File(absolutePath);
        if (!binFile.exists()) {
            throw new RuntimeException("Failed to find replicator binary at path: " + absolutePath);
        }
        if (!binFile.canExecute()) {
            log.info("Setting binary file as executable");
            boolean canExecute = binFile.setExecutable(true);
            if (!canExecute) {
                throw new RuntimeException("Failed to set execute permissions on " + binFile.getAbsolutePath());
            }
        }
    }

    private void createHome() {
        FileUtils.createDirectory(jfProductHomeDir, true);
        replicatorBinDir = new File(jfProductHomeDir.getPath() + "/app/replicator/bin");
        FileUtils.createDirectory(replicatorBinDir, true);
        replicatorEtcDir = new File(jfProductHomeDir.getPath() + "/var/etc/replicator");
        FileUtils.createDirectory(replicatorEtcDir, true);
    }

    public void stop() {
        this.process.getProcess().destroy();
    }

}

