package org.artifactory.lifecycle.storage.db.init;

import org.artifactory.addon.ArtifactoryRunningMode;
import org.artifactory.common.home.ArtifactoryHome;
import org.artifactory.converter.ConverterManager;
import org.artifactory.state.ArtifactoryServerState;
import org.artifactory.state.ArtifactoryStateManagerUtil;
import org.artifactory.storage.db.servers.service.ArtifactoryServersCommonService;
import org.artifactory.version.CompoundVersionDetails;
import org.artifactory.version.VersionProvider;
import org.jfrog.config.ifc.ConfigurationManager;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import static org.artifactory.addon.ArtifactoryRunningMode.HA;
import static org.artifactory.addon.ArtifactoryRunningMode.PRO;

/**
 * A post-DB schema initialization manager.
 *
 * @author Uriah Levy
 */
public class PostDbSchemaInitializationManager {
    private static final Logger log = LoggerFactory.getLogger(PostDbSchemaInitializationManager.class);
    public final ArtifactoryHome artifactoryHome;
    public final ConfigurationManager configurationManager;
    private final VersionProvider versionProvider;
    private final ConverterManager convertersManager;
    private final ArtifactoryServersCommonService serversCommonService;
    private boolean newInstallation;
    private HaInitLock haInitLock;

    public PostDbSchemaInitializationManager(ArtifactoryHome artifactoryHome,
            ConfigurationManager configurationManager, VersionProvider versionProvider,
            ConverterManager convertersManager,
            ArtifactoryServersCommonService serversCommonService,
            boolean newInstallation, HaInitLock haInitLock) {
        this.artifactoryHome = artifactoryHome;
        this.configurationManager = configurationManager;
        this.versionProvider = versionProvider;
        this.convertersManager = convertersManager;
        this.serversCommonService = serversCommonService;
        this.newInstallation = newInstallation;
        this.haInitLock = haInitLock;
    }

    public void init() {
        log.info("Initializing Post-DB initialization manager");
        // Wrap the first init phase inside a lock, and don't release it yet
        haInitLock.runInsideInitLockIfNeeded(this::convertAndInit, this::shouldAcquireInitLock);
        log.info("Post-DB initialization manager initialized");
    }

    boolean shouldAcquireInitLock() {
        CompoundVersionDetails versionDetails = versionProvider.getCompoundVersionDetailsFromDbIfExists();
        return artifactoryHome.isHaConfigured() &&
                (newInstallation || versionDetails == null || !versionDetails.isCurrent());
    }

    private void convertAndInit() {
        convertersManager.convertDatabase();
        // Create the initial server state in this early stage in order to prevent an orphan HA init lock cleanup
        createInitialServerState();
        convertersManager.convertMarkers();
        convertersManager.convertHome();
        convertersManager.convertHomeSync();
        configurationManager.startSync();
    }

    private void createInitialServerState() {
        ArtifactoryStateManagerUtil
                .createOrUpdateArtifactoryServerState(artifactoryHome.getServerId(), ArtifactoryServerState.STARTING,
                        getRunningMode(artifactoryHome), versionProvider.getRunning(),
                        serversCommonService);
    }

    private ArtifactoryRunningMode getRunningMode(ArtifactoryHome artifactoryHome) {
        return artifactoryHome.isHaConfigured() ? HA : PRO;
    }
}
