/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2018 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.storage.db.event.dao;

import com.google.common.collect.Lists;
import com.google.common.collect.Sets;
import org.artifactory.common.ConstantValues;
import org.artifactory.common.config.db.ArtifactoryDbProperties;
import org.artifactory.repo.RepoPathFactory;
import org.artifactory.storage.db.DbService;
import org.artifactory.storage.db.event.entity.EventRecord;
import org.artifactory.storage.db.event.entity.EventsDaoParams;
import org.artifactory.storage.db.util.BaseDao;
import org.artifactory.storage.db.util.JdbcHelper;
import org.artifactory.storage.db.util.querybuilder.ArtifactoryQueryWriter;
import org.artifactory.storage.event.EventType;
import org.jetbrains.annotations.NotNull;
import org.jfrog.common.ResultSetStream;
import org.jfrog.storage.util.DbUtils;
import org.jfrog.storage.util.querybuilder.QueryWriter;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Repository;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.StringUtils;

import javax.annotation.Nonnull;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.stream.Stream;

/**
 * A data access object for the events table.
 *
 * @author Yossi Shaul
 */
@Repository
public class EventsDao extends BaseDao {

    private static final Logger log = LoggerFactory.getLogger(EventsDao.class);

    /**
     * Delete range on node_events which doesn't have reference in node_events_errors
     */
    private static final String DELETE_RANGE_SQL = "DELETE FROM node_events " +
            "WHERE timestamp >= ? AND timestamp < ? " +
            "AND event_id NOT IN " +
            "(SELECT event_id " +
            "FROM node_events_errors " +
            "WHERE event_id IS NOT NULL AND event_time >= ? AND event_time < ?)";

    private static final String INSERT_EVENT_SQL = "INSERT INTO node_events " +
            "(%1$s timestamp, event_type, path, hints) " +
            "VALUES (%2$s ?, ?, ?, ?)";
    private static final String DELETE_TEMP_EVENTS_SQL = "DELETE FROM node_events_tmp " +
            "WHERE tmp_event_id IN (#)";
    public static final int TMP_EVENTS_DELETE_BATCH_SIZE = 100;
    private static final String EVENT_ID_COLUMN = "event_id,";
    private static final String SEQ_ORA = "NEXTVAL('node_events_event_id_seq'),";
    private static final String SEQ_PG = "node_events_event_id_seq.NEXTVAL,";
    private static final String SEQ_DERBY = "NEXT VALUE FOR node_events_event_id_seq,";

    private final DbService dbService;
    private final String insertEventSql;
    private final String selectTmpEventsSql;

    @Autowired
    public EventsDao(JdbcHelper jdbcHelper, ArtifactoryDbProperties dbProperties, DbService dbService) {
        super(jdbcHelper);
        this.dbService = dbService;

        this.selectTmpEventsSql = new QueryWriter(dbProperties.getDbType())
                .select("timestamp, event_type, path, hints, tmp_event_id")
                .from("node_events_tmp")
                .limit(ConstantValues.shiftEventsSqlLimit.getLong())
                .orderBy("timestamp, tmp_event_id")
                .build();

        switch (dbProperties.getDbType()) {
            case POSTGRESQL:
                this.insertEventSql =
                        String.format(INSERT_EVENT_SQL, EVENT_ID_COLUMN, SEQ_ORA);
                break;
            case ORACLE:
                this.insertEventSql =
                        String.format(INSERT_EVENT_SQL, EVENT_ID_COLUMN, SEQ_PG);
                break;
            case DERBY:
                this.insertEventSql =
                        String.format(INSERT_EVENT_SQL, EVENT_ID_COLUMN, SEQ_DERBY);
                break;
            default:
                this.insertEventSql = String.format(INSERT_EVENT_SQL, "", "");
        }
    }

    public void create(EventRecord e) throws SQLException {
        jdbcHelper.executeUpdate("INSERT INTO node_events_tmp (tmp_event_id, timestamp, event_type, path, hints) " +
                        "VALUES(?, ?, ?, ?, ?)",
                dbService.nextId(), e.getTimestamp(), e.getType().code(), e.getPath(), e.getHints());
    }

    /**
     * Loads all the events newer from the given timestamp
     *
     * @param timestamp should be replaced to eventId
     * @return Events newer than the given timestamp ordered by timestamp
     */
    @Nonnull
    @Deprecated
    public List<EventRecord> loadNewer(long timestamp, String repoKey) throws SQLException {
        return loadEventsSinceTimestamp(timestamp, repoKey, false);
    }

    /**
     * Loads all the events newer or same as the given eventId
     *
     * @param timestamp timestamp of the event
     * @return Events newer than the given eventId ordered by eventId
     */
    @Deprecated
    public List<EventRecord> loadNewerOrSameTimestamp(long timestamp, String repoKey) throws SQLException {
        return loadEventsSinceTimestamp(timestamp, repoKey, true);
    }

    public List<EventRecord> loadNewerOrSame(long eventId) throws SQLException {
        return loadEventsSince(eventId, true);
    }

    @Nonnull
    public List<EventRecord> loadEventsByTimestamp(long timestamp) throws SQLException {
        ResultSet rs = null;
        try {
            List<EventRecord> entries = Lists.newArrayList();
            rs = jdbcHelper.executeSelect("SELECT * FROM node_events " +
                    "WHERE timestamp = ? ORDER BY timestamp ASC, event_id ASC", timestamp);
            while (rs.next()) {
                entries.add(eventFromResultSet(rs));
            }
            return entries;
        } finally {
            DbUtils.close(rs);
        }
    }

    @Nonnull
    private List<EventRecord> loadEventsSinceTimestamp(long timestamp, String repoKey, boolean inclusive)
            throws SQLException {
        return loadEventsByParams(EventsDaoParams.builder()
                .timestamp(timestamp)
                .inclusive(inclusive)
                .repoKey(repoKey)
                .build());
    }

    @Nonnull
    private List<EventRecord> loadEventsSince(long eventId, boolean inclusive) throws SQLException {
        return loadEventsByParams(EventsDaoParams.builder()
                .eventId(eventId)
                .inclusive(inclusive)
                .build());
    }

    @Nonnull
    public List<EventRecord> loadEventsByParams(EventsDaoParams queryParams) throws SQLException {
        try (ResultSet rs = getResultSet(queryParams, false)) {
            List<EventRecord> entries = Lists.newArrayList();
            while (rs.next()) {
                entries.add(eventFromResultSet(rs));
            }
            return entries;
        }
    }

    public int countEventsByParams(EventsDaoParams queryParams) throws SQLException {
        if (queryParams.getLimit() > 0) {
            throw new IllegalArgumentException("limit is not allowed");
        }
        try (ResultSet resultSet = getResultSet(queryParams, true)) {
            int count = 0;
            if (resultSet.next()) {
                count = resultSet.getInt(1);
            }
            return count;
        }
    }


    /**
     * Loads all the events newer from the fromTimeStamp and older than the toTimestamp
     *
     * @param fromTimestamp Beginning of the interval non-inclusive timestamp in millis
     * @param toTimestamp   End of the interval non-inclusive timestamp in millis
     * @return Events newer than the given fromTimestamp and older then the toTimestamp ordered by timestamp
     */
    @Nonnull
    public Stream<EventRecord> loadIntervalStream(long fromTimestamp, long toTimestamp, String repoKey, long limit,
            CompletableFuture<SQLException> onException) {
        QueryWriter queryWriter = new ArtifactoryQueryWriter();
        queryWriter.select().from("node_events")
                .where("timestamp > ? AND timestamp < ? AND path like ?")
                .limit(limit)
                .orderBy("timestamp ASC, event_id ASC");
        try {
            ResultSet rs = jdbcHelper.executeSelect(queryWriter.build(), fromTimestamp, toTimestamp, repoKey + "/%");
            return ResultSetStream.asStream(rs, this::eventFromResultSet, onException);
        } catch (SQLException e) {
            onException.complete(e);
            return Stream.empty();
        }
    }

    /**
     * Return max events by timestamp
     */
    @Nonnull
    public Stream<EventRecord> loadNewerStream(long timestamp, long maxEvents,
            CompletableFuture<SQLException> onException) {
        try {
            QueryWriter queryWriter = new ArtifactoryQueryWriter();
            String sql = queryWriter.select().from("node_events")
                    .where("timestamp > ?")
                    .orderBy("timestamp ASC, event_id ASC")
                    .limit(maxEvents)
                    .build();
            ResultSet rs = jdbcHelper.executeSelect(sql, timestamp);
            return ResultSetStream.asStream(rs, this::eventFromResultSet, onException);
        } catch (SQLException e) {
            onException.complete(e);
            return Stream.empty();
        }
    }

    public long getEventsCount() throws SQLException {
        return jdbcHelper.executeSelectCount("SELECT COUNT(*) FROM node_events");
    }

    public long getFirstEventTimestamp() throws SQLException {
        return jdbcHelper.executeSelectLong("SELECT MIN(timestamp) FROM node_events");
    }

    public long getLastEventTimestamp() throws SQLException {
        return jdbcHelper.executeSelectLong("SELECT MAX(timestamp) FROM node_events");
    }

    public long getFirstEventId(String repoKey) throws SQLException {
        return jdbcHelper.executeSelectLong("SELECT MIN(event_id) FROM node_events WHERE path LIKE ?", repoKey + "/%");
    }

    public long getLastEventId() throws SQLException {
        return jdbcHelper.executeSelectLong("SELECT MAX(event_id) FROM node_events");
    }

    public long getOldestTimestampByEventIdInclusive(long eventId) throws SQLException {
        //This method is used to get the timestamp of the min eventId during events cleanup.
        // We might have different order for event ids and timestamp
        // EventId, Timestamp
        // 100,     200
        // 101,     199
        // Min EventId = 100, The timestamp is 200, we will delete everything until timestamp 200
        // including event 101
        // The bellow query returns the min timestamp for all the events equals and above the specified event id
        return jdbcHelper.executeSelectLong("SELECT MIN(timestamp) FROM node_events WHERE event_id >= ?", eventId);
    }

    public List<EventRecord> loadAll() throws SQLException {
        ResultSet rs = null;
        try {
            List<EventRecord> entries = Lists.newArrayList();
            rs = jdbcHelper.executeSelect("SELECT * FROM node_events ORDER BY event_id ASC");
            while (rs.next()) {
                entries.add(eventFromResultSet(rs));
            }
            return entries;
        } finally {
            DbUtils.close(rs);
        }
    }

    public Stream<EventRecord> loadAllStream(CompletableFuture<SQLException> onException) throws SQLException {
        return ResultSetStream.asStream(
                jdbcHelper.executeSelect("SELECT * FROM node_events ORDER BY timestamp ASC, event_id ASC"),
                this::eventFromResultSet,
                onException);
    }

    @NotNull
    private ResultSet getResultSet(EventsDaoParams queryParams, boolean count) throws SQLException {
        boolean inclusive = queryParams.isInclusive();
        long eventId = queryParams.getEventId();
        long timestamp = queryParams.getTimestamp();
        String suffix = queryParams.getSuffix();
        String repoKey = queryParams.getRepoKey();
        Integer limit = !count && queryParams.getLimit() > 0 ? queryParams.getLimit() : null;
        boolean usePath = StringUtils.hasLength(repoKey) || StringUtils.hasLength(suffix);

        //prepare where clause
        //whether to include or not events with the same timestamp
        String comparator = inclusive ? ">=" : ">";
        //use timestamp field only if requested by the params.
        String fieldNameToCompare = timestamp == Long.MIN_VALUE ? "event_id " : "timestamp ";
        String whereClause = fieldNameToCompare + comparator + " ? " + (usePath ? "AND path like ?" : "");

        //prepare the sql
        String sql = new ArtifactoryQueryWriter().from("node_events")
                .select(count ? "count(1)" : "*")
                .where(whereClause)
                .limit(limit != null ? limit.longValue() : null)
                .orderBy(count ? null : "event_id ASC").build();

        long valueToCompare = timestamp == Long.MIN_VALUE ? eventId : timestamp;
        if (usePath) {
            String pathSuffix = suffix == null ? "" : suffix;
            String repoKeyPrefix = StringUtils.hasLength(repoKey) ? repoKey + "/" : "";
            return jdbcHelper.executeSelect(sql, valueToCompare, repoKeyPrefix + "%" + pathSuffix);
        } else {
            return jdbcHelper.executeSelect(sql, valueToCompare);
        }
    }

    /**
     * Delete all events from the event log. Internal use only!
     */
    public void deleteAll() throws SQLException {
        jdbcHelper.executeUpdate("DELETE FROM node_events_errors");
        jdbcHelper.executeUpdate("DELETE FROM node_events_tmp");
        jdbcHelper.executeUpdate("DELETE FROM node_events");
    }

    /**
     * Delete all events from in range.
     */
    public void deleteRange(long startInclusive, long endExclusive) throws SQLException {
        jdbcHelper.executeUpdate(DELETE_RANGE_SQL, startInclusive, endExclusive,
                startInclusive, endExclusive);
    }

    private EventRecord eventFromResultSet(ResultSet rs) throws SQLException {
        return new EventRecord(rs.getLong("event_id"),
                rs.getLong("timestamp"),
                EventType.fromCode(rs.getShort("event_type")),
                rs.getString("path"),
                rs.getString("hints"));
    }

    /**
     * Shift events from node_events_tmp to node_events And delete the tmp records.
     * @return the effected repos
     * @throws SQLException
     */
    @Transactional
    public Set<String> shiftAndDeleteTmpEvents() throws SQLException {
        int copied = 0;
        int deleted = 0;
        Set<String> reposKey = Sets.newHashSet();
        List<List<Long>> idsPartitions = new ArrayList<>();
        List<List<Object>> records = new ArrayList<>();
        Instant fetchStart = Instant.now();
        try (ResultSet rs = jdbcHelper.executeSelect(selectTmpEventsSql)) {
            while (rs.next()) {
                records.add(Arrays.asList(rs.getLong("timestamp"),
                        rs.getInt("event_type"),
                        rs.getString("path"),
                        rs.getString("hints")));
                List<Long> idsPartition = partition(idsPartitions);
                reposKey.add(RepoPathFactory.create(rs.getString("path")).getRepoKey());
                idsPartition.add(rs.getLong("tmp_event_id"));
                copied++;
            }
        }
        Duration fetchDuration = Duration.between(fetchStart, Instant.now());

        Instant copyStart = Instant.now();
        for (List<Object> record : records) {
            jdbcHelper.executeUpdate(insertEventSql,
                            record.get(0),
                            record.get(1),
                            record.get(2),
                            record.get(3));
        }
        Duration copyDuration = Duration.between(copyStart, Instant.now());

        Instant deleteStart = Instant.now();
        for (List<Long> idsPartition : idsPartitions) {
            deleted += jdbcHelper.executeUpdate(DELETE_TEMP_EVENTS_SQL, idsPartition);
        }
        Duration deleteDuration = Duration.between(deleteStart, Instant.now());

        log.debug("Shift tmp events: shifted={}:(fetch {} ms, copy {} ms), deleted={}:({} ms), repos={}", copied,
                fetchDuration.toMillis(), copyDuration.toMillis(), deleted, deleteDuration.toMillis(), reposKey);
        return reposKey;
    }

    @Nonnull
    List<Long> partition(@Nonnull List<List<Long>> partitions) {
        List<Long> partition = !partitions.isEmpty() ? partitions.get(partitions.size() - 1) : null;
        if (partition == null || partition.size() == TMP_EVENTS_DELETE_BATCH_SIZE) {
            partition = new ArrayList<>(TMP_EVENTS_DELETE_BATCH_SIZE);
            partitions.add(partition);
        }
        return partition;

    }
}
