package org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.fetch;

import com.google.common.collect.Iterables;
import org.artifactory.api.rest.common.model.continues.ContinueResult;
import org.artifactory.api.rest.common.model.continues.FetchFunction;
import org.artifactory.descriptor.repo.FederatedRepoDescriptor;
import org.artifactory.descriptor.repo.LocalCacheRepoDescriptor;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RemoteRepoDescriptor;
import org.artifactory.descriptor.repo.VirtualRepoDescriptor;
import org.artifactory.descriptor.repo.distribution.DistributionRepoDescriptor;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.TreeFilter;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.RepositoryNode;
import org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.repo.VirtualRemoteRepositoryNode;
import org.artifactory.util.CollectionUtils;
import org.glassfish.jersey.internal.guava.Lists;
import org.jfrog.common.StreamSupportUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

import static org.artifactory.repo.RepoDetailsType.DISTRIBUTION;
import static org.artifactory.repo.RepoDetailsType.LOCAL;
import static org.artifactory.repo.RepoPath.REMOTE_CACHE_SUFFIX;
import static org.artifactory.ui.rest.model.artifacts.browse.treebrowser.nodes.RepositoryType.CACHED;

/**
 * @author Omri Ziv
 */
public class RootFetchByRepositoryType extends RootFetchStrategy {

    private static final Logger log = LoggerFactory.getLogger(RootFetchByRepositoryType.class);

    private Collection<RepositoryType> repoOrder;
    private final Map<RepositoryType, FetchFunction<RepositoryNode>> repoDetailsTypeFetchFunctionMap;

    public RootFetchByRepositoryType(TreeFilter treeFilter, Collection<RepositoryType> repoOrder) {
        super(treeFilter);
        this.repoDetailsTypeFetchFunctionMap = buildMap();
        this.repoOrder = repoOrder;
    }

    private Map<RepositoryType, FetchFunction<RepositoryNode>> buildMap() {
        Map<RepositoryType, FetchFunction<RepositoryNode>> map = new HashMap<>();
        map.put(RepositoryType.DISTRIBUTION, new FetchFunction<>(this::getDistributionRepoNodes,
                (long) repositoryService.getDistributionRepoDescriptorsByProject(projectKey).size()));
        map.put(RepositoryType.LOCAL, new FetchFunction<>(this::getLocalRepoNodes,
                (long) repositoryService.getLocalRepoDescriptorsIncludingInfoReposByProject(projectKey).size()));
        map.put(RepositoryType.FEDERATED, new FetchFunction<>(this::getFederatedRepoNodes,
                (long) repositoryService.getFederatedRepoDescriptorsByProject(projectKey).size()));
        map.put(CACHED, new FetchFunction<>(this::getCachedRepoNodes,
                (long) repositoryService.getCachedRepoDescriptorsByProject(projectKey).size()));
        map.put(RepositoryType.REMOTE, new FetchFunction<>(this::getRemoteRepoNodes,
                (long) repositoryService.getRemoteRepoDescriptorsByProject(projectKey).size()));
        map.put(RepositoryType.VIRTUAL, new FetchFunction<>(this::getVirtualRepoNodes,
                (long) repositoryService.getVirtualRepoDescriptorByProject(projectKey).size()));

        return map;
    }

    @Override
    protected List<FetchFunction<RepositoryNode>> getFetchFunctions() {
        List<FetchFunction<RepositoryNode>> superFunctions = super.getFetchFunctions();
        List<RepositoryType> repositoryTypes = treeFilter.getRepositoryTypes();
        boolean emptyRepos = CollectionUtils.isNullOrEmpty(repositoryTypes);
        List<FetchFunction<RepositoryNode>> fetchFunctions = StreamSupportUtils.stream(this.repoOrder)
                .filter(repoType -> emptyRepos || repositoryTypes.contains(repoType))
                .map(repoDetailsTypeFetchFunctionMap::get)
                .collect(Collectors.toList());

        List<FetchFunction<RepositoryNode>> allFunctions =
                Lists.newArrayList(Iterables.concat(superFunctions, fetchFunctions));
        log.debug("{} Candidates Fetch Function", allFunctions.size());
        return allFunctions;

    }

    private ContinueResult<RepositoryNode> getDistributionRepoNodes(Integer skip, Integer limit) {
        List<DistributionRepoDescriptor> originalSortedDistributionDescriptors =
                sortAndGet(repositoryService.getDistributionRepoDescriptorsByProject(projectKey));

        List<DistributionRepoDescriptor> distributionDescriptors =
                StreamSupportUtils.stream(originalSortedDistributionDescriptors)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());

        int lastIndex = getLastIndex(originalSortedDistributionDescriptors, distributionDescriptors, limit);
        return new ContinueResult<>(lastIndex, getDistributionNodes(distributionDescriptors));
    }

    private ContinueResult<RepositoryNode> getLocalRepoNodes(Integer skip, Integer limit) {
        List<LocalRepoDescriptor> originalSortedLocalRepos =
                sortAndGet(repositoryService.getLocalRepoDescriptorsIncludingInfoReposByProject(projectKey));

        List<LocalRepoDescriptor> localRepoDescriptors =
                StreamSupportUtils.stream(originalSortedLocalRepos)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());
        int lastIndex = getLastIndex(originalSortedLocalRepos, localRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getLocalNodes(localRepoDescriptors));
    }

    private ContinueResult<RepositoryNode> getFederatedRepoNodes(Integer skip, Integer limit) {
        List<FederatedRepoDescriptor> originalSortedFederatedRepos =
                sortAndGet(repositoryService.getFederatedRepoDescriptorsByProject(projectKey));

        List<FederatedRepoDescriptor> localRepoDescriptors =
                StreamSupportUtils.stream(originalSortedFederatedRepos)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());
        int lastIndex = getLastIndex(originalSortedFederatedRepos, localRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getLocalNodes(localRepoDescriptors));
    }

    private ContinueResult<RepositoryNode> getCachedRepoNodes(Integer skip, Integer limit) {
        List<LocalCacheRepoDescriptor> originalSortedLocalRepos =
                sortAndGet(repositoryService.getCachedRepoDescriptorsByProject(projectKey));

        List<LocalCacheRepoDescriptor> localRCachedepoDescriptors =
                StreamSupportUtils.stream(originalSortedLocalRepos)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());
        int lastIndex = getLastIndex(originalSortedLocalRepos, localRCachedepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getLocalNodes(localRCachedepoDescriptors));
    }

    private ContinueResult<RepositoryNode> getRemoteRepoNodes(Integer skip, Integer limit) {
        List<RemoteRepoDescriptor> originalSortedRemoteDescriptors =
                sortAndGet(repositoryService.getRemoteRepoDescriptorsByProject(projectKey));

        List<RemoteRepoDescriptor> remoteRepoDescriptors =
                StreamSupportUtils.stream(originalSortedRemoteDescriptors)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());
        int lastIndex = getLastIndex(originalSortedRemoteDescriptors, remoteRepoDescriptors, limit);
        return new ContinueResult<>(lastIndex, getRemoteNodes(remoteRepoDescriptors));
    }

    /**
     * add virtual repo nodes to repo list
     */
    private ContinueResult<RepositoryNode> getVirtualRepoNodes(Integer skip, Integer limit) {
        List<VirtualRepoDescriptor> originalSortedVirtualDescriptors =
                sortAndGet(repositoryService.getVirtualRepoDescriptorByProject(projectKey));

        List<VirtualRepoDescriptor> virtualDescriptors =
                StreamSupportUtils.stream(originalSortedVirtualDescriptors)
                        .skip(skip)
                        .filter(this::accept)
                        .limit(limit)
                        .collect(Collectors.toList());
        int lastIndex = getLastIndex(originalSortedVirtualDescriptors, virtualDescriptors, limit);
        return new ContinueResult<>(lastIndex, getVirtualNodes(virtualDescriptors));
    }


    private List<RepositoryNode> getLocalNodes(List<? extends LocalRepoDescriptor> repos) {
        return StreamSupportUtils.stream(repos)
                .map(repo -> {
                    String repoType =
                            repo.getKey().endsWith(REMOTE_CACHE_SUFFIX) ? CACHED.getTypeNameLoweCase() :
                                    LOCAL.typeNameLowercase();
                    return new RepositoryNode(repo.getKey(), repo.getType(), repoType);
                })
                .collect(Collectors.toList());
    }

    private List<RepositoryNode> getRemoteNodes(List<RemoteRepoDescriptor> repos) {
        return StreamSupportUtils.stream(repos)
                .map(VirtualRemoteRepositoryNode::fromRemoteDescriptor)
                .collect(Collectors.toList());
    }

    private List<RepositoryNode> getVirtualNodes(List<VirtualRepoDescriptor> repos) {
        return StreamSupportUtils.stream(repos)
                .map(VirtualRemoteRepositoryNode::fromVirtualDescriptor)
                .collect(Collectors.toList());
    }

    private List<RepositoryNode> getDistributionNodes(List<DistributionRepoDescriptor> repos) {
        return StreamSupportUtils.stream(repos)
                .map(repo -> new RepositoryNode(repo.getKey(), repo.getType(), DISTRIBUTION.typeNameLowercase()))
                .collect(Collectors.toList());
    }
}
