/*
 *
 * Artifactory is a binaries repository manager.
 * Copyright (C) 2020 JFrog Ltd.
 *
 * Artifactory is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 * Artifactory is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with Artifactory.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package org.artifactory.ui.rest.service.admin.security.keys.keypairs;

import org.artifactory.addon.common.keys.KeyPairService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.repo.KeyPairDescriptor;
import org.artifactory.descriptor.repo.RepoBaseDescriptor;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.RestResponse;
import org.artifactory.rest.common.service.RestService;
import org.artifactory.ui.rest.model.admin.security.keys.KeyPairIModel;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import java.util.List;
import java.util.Map;
import java.util.stream.Collectors;

/**
 * @author Yuval Reches
 */
@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
public class GetAllKeyPairsService implements RestService {

    private final CentralConfigService centralConfigService;
    private final KeyPairService keyPairService;

    @Autowired
    public GetAllKeyPairsService(CentralConfigService centralConfigService, KeyPairService keyPairService) {
        this.centralConfigService = centralConfigService;
        this.keyPairService = keyPairService;
    }

    @Override
    public void execute(ArtifactoryRestRequest request, RestResponse response) {
        List<String> types = request.getQueryParamListByKey("type");
        List<KeyPairDescriptor> keyPairs = keyPairService.getKeyPairs(types);
        response.iModelList(keyPairs.stream().map(this::createKeyPairIModel).collect(Collectors.toList()));
    }

    private KeyPairIModel createKeyPairIModel(KeyPairDescriptor keyPair) {
        KeyPairIModel keyPairIModel = new KeyPairIModel();
        keyPairIModel.setPairName(keyPair.getKey());
        keyPairIModel.setAlias(keyPair.getAlias());
        keyPairIModel.setResources(getListOfReposPerKeyPair(keyPair.getKey()));
        keyPairIModel.setResourcesAsSecondary(getListOfReposPerKeyPairAsSecondary(keyPair.getKey()));
        return keyPairIModel;
    }

    List<String> getListOfReposPerKeyPair(String key) {
        List<String> localReposList =
                getReposWithPair(key, centralConfigService.getMutableDescriptor().getLocalRepositoriesMap());
        List<String> federatedRepoList =
                getReposWithPair(key, centralConfigService.getMutableDescriptor().getFederatedRepositoriesMap());
        List<String> virtualReposList =
                getReposWithPair(key, centralConfigService.getMutableDescriptor().getVirtualRepositoriesMap());
        localReposList.addAll(federatedRepoList);
        localReposList.addAll(virtualReposList);
        return localReposList;
    }

    private List<String> getListOfReposPerKeyPairAsSecondary(String key) {
        List<String> localReposList = getReposWithPairAsSecondary(key,
                        centralConfigService.getMutableDescriptor().getLocalRepositoriesMap());
        List<String> virtualReposList = getReposWithPairAsSecondary(key,
                        centralConfigService.getMutableDescriptor().getVirtualRepositoriesMap());
        localReposList.addAll(virtualReposList);
        return localReposList;
    }

    private List<String> getReposWithPair(String key, Map<String, ? extends RepoBaseDescriptor> reposList) {
        return reposList.values().stream()
                .filter(repo -> repo.getPrimaryKeyPairDescriptor() != null)
                .filter(repo -> key.equals(repo.getPrimaryKeyPairDescriptor().getKey()))
                .map(RepoBaseDescriptor::getKey).collect(Collectors.toList());
    }

    private List<String> getReposWithPairAsSecondary(String key, Map<String, ? extends RepoBaseDescriptor> reposList) {
        return reposList.values().stream()
                .filter(repo -> repo.getSecondaryKeyPairDescriptor() != null)
                .filter(repo -> key.equals(repo.getSecondaryKeyPairDescriptor().getKey()))
                .map(RepoBaseDescriptor::getKey).collect(Collectors.toList());
    }

}