package org.artifactory.ui.rest.service.admin.configuration.repositories;

import org.apache.commons.lang.StringUtils;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.descriptor.config.CentralConfigDescriptorImpl;
import org.artifactory.projects.ProjectsService;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.rest.common.exception.RepoConfigException;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.ui.rest.model.admin.configuration.repository.GeneralRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.federation.FederatedRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.CommonLocalRepositoryConfigModel;
import org.artifactory.ui.rest.model.admin.configuration.repository.local.LocalRepositoryConfigModel;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.CreateRepoConfigHelper;
import org.artifactory.ui.rest.service.admin.configuration.repositories.util.validator.RepoConfigValidator;
import org.mockito.ArgumentCaptor;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import static org.mockito.ArgumentMatchers.eq;
import static org.mockito.Mockito.never;
import static org.mockito.Mockito.spy;
import static org.mockito.Mockito.verify;
import static org.testng.AssertJUnit.assertEquals;
import static org.testng.AssertJUnit.assertNotNull;
import static org.testng.AssertJUnit.fail;

/**
 * @author AndreiK.
 */
@Test
public class CreateRepositoryConfigServiceTest {

    @Mock
    CentralConfigService centralConfigService;
    @Mock
    RepoConfigValidator repoConfigValidator;
    @Mock
    CreateRepoConfigHelper createRepoConfigHelper;
    @Mock
    ProjectsService projectsService;

    CreateRepositoryConfigService service;

    @BeforeMethod
    public void setup() {
        MockitoAnnotations.openMocks(this);
        service = new CreateRepositoryConfigService(centralConfigService, repoConfigValidator, createRepoConfigHelper,
                projectsService);
    }

    @Test(dataProvider = "repoTypesProvider")
    public void testCreateRepo(String modelClass, String repoType) throws RepoConfigException {
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        CommonLocalRepositoryConfigModel model = null;
        if (LocalRepositoryConfigModel.class.getName().equals(modelClass)) {
            model = spy(new LocalRepositoryConfigModel());
        } else if (FederatedRepositoryConfigModel.class.getName().equals(modelClass)) {
            model = spy(new FederatedRepositoryConfigModel());
        }
        GeneralRepositoryConfigModel generalModel = new GeneralRepositoryConfigModel();
        generalModel.setRepoKey(repoType);
        assertNotNull("Unrecognized class name was provided", model);
        model.setGeneral(generalModel);

        service.createRepo(response, model, descriptor);

        ArgumentCaptor<String> repoTypeCaptor = ArgumentCaptor.forClass(String.class);
        verify(repoConfigValidator).validateRepoName(eq(repoType), repoTypeCaptor.capture());
        assertEquals(repoType, repoTypeCaptor.getValue());
        verify(model).createRepo(createRepoConfigHelper, descriptor);
    }

    @Test(dataProvider = "repoModelWithNameContainingDashCacheProvider")
    public void testCreateRepoFailOnDashCash(String modelClass, String repoKey) throws RepoConfigException {
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        CentralConfigDescriptorImpl descriptor = new CentralConfigDescriptorImpl();
        CommonLocalRepositoryConfigModel model = null;
        if (LocalRepositoryConfigModel.class.getName().equals(modelClass)) {
            model = spy(new LocalRepositoryConfigModel());
        } else if (FederatedRepositoryConfigModel.class.getName().equals(modelClass)) {
            model = spy(new FederatedRepositoryConfigModel());
        }
        GeneralRepositoryConfigModel generalModel = new GeneralRepositoryConfigModel();
        generalModel.setRepoKey(repoKey);
        assertNotNull("Unrecognized class name was provided", model);
        model.setGeneral(generalModel);

        try {
            service.createRepo(response, model, descriptor);
            fail("Expected exception to be thrown");
        } catch (RepoConfigException e) {
            ArgumentCaptor<String> repoTypeCaptor = ArgumentCaptor.forClass(String.class);
            verify(repoConfigValidator).validateRepoName(eq(repoKey), repoTypeCaptor.capture());
            Assert.assertEquals(e.getMessage(),
                    StringUtils.capitalize(repoTypeCaptor.getValue()) + " repository '" + repoKey + "' with " +
                            "'-cache' suffix is not allowed");
            String configModelRepoType = service.getRepoTypeByConfigModel(model);
            assertEquals(configModelRepoType, repoTypeCaptor.getValue());
            verify(model, never()).createRepo(createRepoConfigHelper, descriptor);
        }
    }

    @DataProvider
    private Object[][] repoTypesProvider() {
        return new Object[][] {
                {LocalRepositoryConfigModel.class.getName(), RepoDetailsType.LOCAL_REPO},
                {FederatedRepositoryConfigModel.class.getName(), RepoDetailsType.FEDERATED_REPO},
        };
    }

    @DataProvider
    private Object[][] repoModelWithNameContainingDashCacheProvider() {
        return new Object[][] {
                {LocalRepositoryConfigModel.class.getName(), RepoDetailsType.LOCAL_REPO + "-cache"},
                {FederatedRepositoryConfigModel.class.getName(), RepoDetailsType.FEDERATED_REPO + "-cache"},
        };
    }
}