package org.artifactory.ui.rest.service.setmeup;

import lombok.NonNull;
import org.apache.commons.lang3.StringUtils;
import org.artifactory.addon.AddonsManager;
import org.artifactory.addon.CoreAddons;
import org.artifactory.aol.AolService;
import org.artifactory.api.config.CentralConfigService;
import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.api.repo.RepositoryService;
import org.artifactory.api.security.AuthorizationService;
import org.artifactory.descriptor.repo.LocalRepoDescriptor;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.projects.ProjectsService;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.rest.common.service.ArtifactoryRestRequest;
import org.artifactory.rest.common.service.ArtifactoryRestResponse;
import org.artifactory.test.ArtifactoryHomeBoundTest;
import org.artifactory.ui.rest.model.setmeup.SetMeUpModel;
import org.artifactory.ui.rest.model.utils.repositories.RepoKeyType;
import org.jetbrains.annotations.NotNull;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.MockitoAnnotations;
import org.testng.Assert;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;
import org.testng.collections.Lists;

import javax.servlet.http.HttpServletRequest;
import java.util.ArrayList;
import java.util.List;

import static org.mockito.ArgumentMatchers.anyString;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

/**
 * @author Alexei Vainshtein
 */
public class GetSetMeUpServiceTest extends ArtifactoryHomeBoundTest {

    private static final String PROJECT_X = "projectX";
    private static final String REPO_A = "repoA";
    private static final String REPO_B = "repoB";

    @Mock
    private CentralConfigService centralConfigService;
    @Mock
    private RepositoryService repositoryService;
    @Mock
    private AuthorizationService authorizationService;
    @InjectMocks
    private GetSetMeUpService setMeUpService;

    @BeforeMethod
    private void setup() {
        MockitoAnnotations.openMocks(this);
    }

    @AfterMethod
    private void cleanup() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test
    public void testSetMeUpModelWithoutFlag() {
        setupMocks(false);
        ArtifactoryRestRequest request = getArtifactoryRestRequest();
        List<RepoKeyType> repoKeyTypes = new ArrayList<>();
        SetMeUpModel setMeUpModel = setMeUpService.getSetMeUpModel(request, repoKeyTypes);
        Assert.assertFalse(setMeUpModel.isSkipDockerDns());
    }

    @Test
    public void testSetMeUpModelWithFlag() {
        setupMocks(true);
        ArtifactoryRestRequest request = getArtifactoryRestRequest();
        List<RepoKeyType> repoKeyTypes = new ArrayList<>();
        SetMeUpModel setMeUpModel = setMeUpService.getSetMeUpModel(request, repoKeyTypes);
        Assert.assertTrue(setMeUpModel.isSkipDockerDns());
    }

    @Test(dataProvider = "provideProjectName")
    public void filterRepositoriesByProjectName(String projectName, boolean hasPermissionOnRepoRoot,
            int expectedReposCount) {
        setupMocks(false);
        ArtifactoryRestRequest request = getArtifactoryRestRequest();
        ArtifactoryRestResponse response = new ArtifactoryRestResponse();
        when(request.getQueryParamByKey(ProjectsService.PROJECT_KEY)).thenReturn(projectName);
        LocalRepoDescriptor defaultRepo = createLocalRepoDescriptor(REPO_A);
        LocalRepoDescriptor projectRepo = createLocalRepoDescriptor(REPO_B);
        when(repositoryService.getLocalAndCachedRepoDescriptorsByProject(ProjectsService.DEFAULT_PROJECT))
                .thenReturn(Lists.newArrayList(defaultRepo, projectRepo));
        when(repositoryService.getLocalAndCachedRepoDescriptorsByProject(PROJECT_X))
                .thenReturn(Lists.newArrayList(projectRepo));
        when(authorizationService.userHasPermissionsOnRepositoryRoot(anyString())).thenReturn(hasPermissionOnRepoRoot);

        //act
        setMeUpService.execute(request, response);
        SetMeUpModel setMeUpModel = (SetMeUpModel) response.getIModel();

        Assert.assertEquals(setMeUpModel.getRepoKeyTypes().size(), expectedReposCount);
    }

    @NotNull
    private LocalRepoDescriptor createLocalRepoDescriptor(@NonNull String repoKey) {
        LocalRepoDescriptor projectRepo = new LocalRepoDescriptor();
        projectRepo.setKey(repoKey);
        projectRepo.setType(RepoType.Generic);
        return projectRepo;
    }

    @NonNull
    private ArtifactoryRestRequest getArtifactoryRestRequest() {
        HttpServletRequest servletRequest = mock(HttpServletRequest.class);
        when(servletRequest.getHeader(ArtifactoryRequest.JFROG_OVERRIDE_BASE_URL))
                .thenReturn("https://test.jfrog.ui/");
        ArtifactoryRestRequest request = mock(ArtifactoryRestRequest.class);
        when(request.getServletRequest()).thenReturn(servletRequest);
        return request;
    }

    private void setupMocks(boolean isSkipDns) {
        ArtifactoryContext artifactoryContext = mock(ArtifactoryContext.class);
        AddonsManager addonsManager = mock(AddonsManager.class);
        CoreAddons coreAddons = mock(CoreAddons.class);
        AolService aolService = mock(AolService.class);
        when(aolService.skipDockerDns()).thenReturn(isSkipDns);
        when(coreAddons.getArtifactoryServerName()).thenReturn("jfrog");
        when(addonsManager.addonByType(CoreAddons.class)).thenReturn(coreAddons);
        when(artifactoryContext.beanForType(AddonsManager.class)).thenReturn(addonsManager);
        when(artifactoryContext.beanForType(AolService.class)).thenReturn(aolService);
        when(centralConfigService.getListeningPort()).thenReturn(-1);
        when(artifactoryContext.beanForType(CentralConfigService.class)).thenReturn(centralConfigService);
        when(artifactoryContext.beanForType(RepositoryService.class)).thenReturn(repositoryService);
        when(artifactoryContext.beanForType(AuthorizationService.class)).thenReturn(authorizationService);
        ArtifactoryContextThreadBinder.bind(artifactoryContext);
    }

    @DataProvider
    public static Object[][] provideProjectName() {
        return new Object[][] {
                {null, true, 2},
                {StringUtils.EMPTY, true, 2},
                {ProjectsService.DEFAULT_PROJECT, true, 2},
                {ProjectsService.DEFAULT_PROJECT, false, 0},
                {PROJECT_X, true, 1},
                {PROJECT_X, false, 0}
        };
    }
}