package org.artifactory.rest.resource.federation;

import org.artifactory.api.security.AuthorizationService;
import org.artifactory.federation.FederatedRepoService;
import org.artifactory.repo.RepoDetails;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.rest.common.util.RestUtils;
import org.artifactory.rest.exception.BadRequestException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.config.BeanDefinition;
import org.springframework.context.annotation.Scope;
import org.springframework.stereotype.Component;

import javax.annotation.security.RolesAllowed;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.PathParam;
import javax.ws.rs.Produces;
import javax.ws.rs.QueryParam;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import java.util.List;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_OK;

/**
 * @author Alexei Vainshtein
 */

@Component
@Scope(BeanDefinition.SCOPE_PROTOTYPE)
@Path("federation")
@RolesAllowed({AuthorizationService.ROLE_ADMIN})
public class FederatedRepoResource {
    private static final Logger log = LoggerFactory.getLogger(FederatedRepoResource.class);

    @Context
    HttpServletRequest httpRequest;

    private FederatedRepoService federatedRepoService;

    @Autowired
    public void setService(FederatedRepoService federatedRepoService) {
        this.federatedRepoService = federatedRepoService;
    }

    @GET
    @Produces({MediaType.APPLICATION_JSON})
    @Path("validate/{repoKey}")
    public Response validate(@PathParam("repoKey") String repoKey, @QueryParam("time") long time) {
        try {
            federatedRepoService.validate(repoKey, httpRequest.getHeader(ArtifactoryRequest.ORIGINAL_CLIENT_URI), time);
            return Response.status(SC_OK).build();
        } catch (BadRequestException e) {
            return Response.status(SC_BAD_REQUEST)
                    .entity("Validation of federation repository failed: " + e.getMessage())
                    .build();
        }
    }

    @GET
    @Produces({MediaType.APPLICATION_JSON})
    @Path("repos")
    public Response repos() {
        // Will return all the federation repos:
        String contextUrl = RestUtils.getServletContextUrl(httpRequest);
        List<RepoDetails> repos = federatedRepoService.getRepos(contextUrl);
        return Response.status(SC_OK).entity(repos).build();
    }

    @POST
    @Path("migrate/{repoKey}")
    @Produces(MediaType.APPLICATION_JSON)
    public Response migrate(@PathParam("repoKey") String repoKey) {
        try {
            federatedRepoService.migrate(repoKey);
        } catch (BadRequestException e) {
            String errorMessage = String.format("Failed migrate repository %s due to: %s", repoKey, e.getMessage());
            log.error(errorMessage, repoKey, e.getMessage());
            log.debug("Failed migrate repository due to:", e);
            return Response.status(SC_BAD_REQUEST).entity(errorMessage).build();
        }
        return Response.status(SC_OK).entity("Migration finished successfully").build();
    }
}
