package org.artifactory.rest.resource.federation;

import org.artifactory.api.context.ArtifactoryContext;
import org.artifactory.api.context.ArtifactoryContextThreadBinder;
import org.artifactory.descriptor.repo.RepoType;
import org.artifactory.federation.FederatedRepoService;
import org.artifactory.repo.RepoDetails;
import org.artifactory.repo.RepoDetailsType;
import org.artifactory.request.ArtifactoryRequest;
import org.artifactory.rest.exception.BadRequestException;
import org.testng.annotations.AfterMethod;
import org.testng.annotations.BeforeMethod;
import org.testng.annotations.DataProvider;
import org.testng.annotations.Test;

import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.core.Response;
import java.util.ArrayList;
import java.util.List;

import static org.apache.http.HttpStatus.SC_BAD_REQUEST;
import static org.apache.http.HttpStatus.SC_OK;
import static org.mockito.Mockito.doAnswer;
import static org.mockito.Mockito.doThrow;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;
import static org.testng.Assert.assertEquals;

/**
 * @author Alexei Vainshtein
 */
public class FederatedRepoResourceTest {

    private FederatedRepoService service;
    private final String path = "path";

    @BeforeMethod
    public void setup() {
        service = mock(FederatedRepoService.class);
        ArtifactoryContext context = mock(ArtifactoryContext.class);
        when(context.getContextPath()).thenReturn(path);
        ArtifactoryContextThreadBinder.bind(context);
    }

    @AfterMethod
    public void tearDown() {
        ArtifactoryContextThreadBinder.unbind();
    }

    @Test(dataProvider = "responseProvider")
    public void testValidate(String responseFromService, int responseCode, String entity) {
        FederatedRepoResource resource = new FederatedRepoResource();
        resource.setService(service);
        createRequest(resource);
        if (responseCode != SC_OK) {
            doThrow(new BadRequestException(responseFromService)).when(service)
                    .validate("repo", "http://localhost:8080/artifactory", 0);
        }
        Response repo = resource.validate("repo", 0);
        assertEquals(repo.getStatus(), responseCode);
        assertEquals(repo.getEntity(), entity);
    }

    @Test
    public void testRepos() {
        FederatedRepoResource resource = new FederatedRepoResource();
        resource.setService(service);
        createRequest(resource);
        List<RepoDetails> repoDetails = new ArrayList<>();
        repoDetails.add(new RepoDetails("key", "", RepoDetailsType.FEDERATED, "url", "", RepoType.Docker));
        repoDetails.add(new RepoDetails("key2", "", RepoDetailsType.FEDERATED, "url", "", RepoType.Npm));
        when(service.getRepos("header/" + path)).thenReturn(repoDetails);
        Response repos = resource.repos();
        assertEquals(repos.getStatus(), SC_OK);
        assertEquals(repos.getEntity(), repoDetails);
    }

    @Test
    public void testMigrate() {
        FederatedRepoResource resource = new FederatedRepoResource();
        resource.setService(service);
        doAnswer(invocationOnMock -> {
            assertEquals(invocationOnMock.getArgument(0, String.class), ("repo"));
            return null;
        }).when(service).migrate("repo");
        Response repos = resource.migrate("repo");
        assertEquals(repos.getStatus(), SC_OK);
        assertEquals(repos.getEntity(), "Migration finished successfully");
    }

    @Test
    public void testMigrateFails() {
        FederatedRepoResource resource = new FederatedRepoResource();
        resource.setService(service);
        doThrow(new BadRequestException("should throw exception test")).when(service).migrate("repo");
        Response repos = resource.migrate("repo");
        assertEquals(repos.getStatus(), SC_BAD_REQUEST);
        assertEquals(repos.getEntity(), "Failed migrate repository repo due to: should throw exception test");
    }

    @DataProvider
    public Object[][] responseProvider() {
        return new Object[][] {
                {null, 200, null},
                {"error", 400, "Validation of federation repository failed: error"},
        };
    }

    private void createRequest(FederatedRepoResource resource) {
        HttpServletRequest request = mock(HttpServletRequest.class);
        when(request.getHeader(ArtifactoryRequest.JFROG_OVERRIDE_BASE_URL)).thenReturn("header");
        when(request.getHeader(ArtifactoryRequest.ORIGINAL_CLIENT_URI)).thenReturn("http://localhost:8080/artifactory");
        resource.httpRequest = request;
    }
}